"""Session module.

The ``Session`` module allows PyEnSight to control the EnSight session.

Examples:

>>> from ansys.pyensight.core import LocalLauncher
>>> session = LocalLauncher().start()
>>> type(session)
ansys.pyensight.Session

"""
import atexit
import importlib.util
from os import listdir
import os.path
import platform
import sys
import textwrap
import time
import types
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union
from urllib.parse import urlparse
from urllib.request import url2pathname
import uuid
import webbrowser

from ansys.pyensight.core.enscontext import EnsContext
from ansys.pyensight.core.launcher import Launcher
from ansys.pyensight.core.listobj import ensobjlist
from ansys.pyensight.core.renderable import (
    RenderableDeepPixel,
    RenderableEVSN,
    RenderableFluidsWebUI,
    RenderableImage,
    RenderableMP4,
    RenderableSGEO,
    RenderableVNC,
    RenderableVNCAngular,
    RenderableWebGL,
)
import requests

if TYPE_CHECKING:
    from ansys.api.pyensight import ensight_api
    from ansys.pyensight.core import enscontext, ensight_grpc, renderable
    from ansys.pyensight.core.ensobj import ENSOBJ
    from ansys.pyensight.core.utils.dsg_server import DSGSession


class InvalidEnSightVersion(Exception):
    pass


class Session:
    """Provides for accessing an EnSight ``Session`` instance.

    The ``Session`` object wraps the various connections to an EnSight instance. It includes
    the location of the installation and the gRPC, HTML and WS ports used to talk to the
    EnSight session. In most cases, a ``Session`` instance is created using Launcher
    class methods, but if the EnSight session is already running, an instance can be
    created directly to wrap this running EnSight session.

    If the ``Session`` object is created via a Launcher ``start()`` method call, when the
    session object is garbage collected, the EnSight instance is automatically stopped.
    To prevent this behavior (and leave the EnSight instance running), set the
    ``halt_ensight_on_close`` property to ``False``.

    A gRPC connection is required to interact with an EnSight session. The host, gRPC
    port number, and secret key must be specified. The HTML and WS ports, which are used to
    enable the :func:`show<ansys.pyensight.core.Session.show>`) method, also require that
    an instance of the websocket server is running.

    Parameters
    ----------
    host : str, optional
        Name of the host that the EnSight gRPC service is running on.
        The default is ``"127.0.0.1"``, which is the localhost.
    install_path : str, optional
        Path to the CEI directory to launch EnSight from.
        The default is ``None``.
    secret_key : str, optional
        Shared session secret key for validating the gRPC communication.
        The default is ``""``.
    grpc_port : int, optional
        Port number of the EnSight gRPC service. The default is ``12345``.
    grpc_use_tcp_sockets : bool, optional
        If using gRPC, and if True, then allow TCP Socket based connections
        instead of only local connections.
    grpc_allow_network_connections : bool, optional
        If using gRPC and using TCP Socket based connections, listen on all networks.
    grpc_disable_tls : bool, optional
        If using gRPC and using TCP Socket based connections, disable TLS.
    grpc_uds_pathname : str, optional
        If using gRPC and using Unix Domain Socket based connections, explicitly
        set the pathname to the shared UDS file instead of using the default.
    html_host : str, optional
        Optional hostname for html connections if different than host
        Used by Ansys Lab and reverse proxy servers
    html_port : int, optional
        Port number of the websocket server's HTTP server. The default is
        ``None``.
    ws_port : int, optional
        Port number of the websocket server's WS server. The default is
        ``None``.
    session_directory : str, optional
        Directory on the server for local data storage. The default is
        ``None``.
    timeout : float, optional
        Number of seconds to try a gRPC connection before giving up.
        The default is ``120``.
    rest_api : bool, optional
        Whether to enable the EnSight REST API for the remote EnSight instance.
        The default is ``False``.
    sos : bool, optional
        Whether the remote EnSight instance is to use the SOS (Server
        of Servers) feature. The default is ``False``.

    Examples
    --------

    >>> from ansys.pyensight.core import Session
    >>> session = Session(host="127.0.0.1", grpc_port=12345, http_port=8000, ws_port=8100)

    >>> from ansys.pyensight.core import LocalLauncher
    >>> session = LocalLauncher().start()

    >>> # Launch an instance of EnSight, then create a second connection to the instance
    >>> from ansys.pyensight.core import LocalLauncher, Session
    >>> launched_session = LocalLauncher().start()
    >>> # Get a string that can be used to create a second connection
    >>> session_string = str(launched_session)
    >>> # Create a second connection to the same EnSight instance
    >>> connected_session = eval(session_string)

    WARNING:
    Overriding the default values for these options: grpc_use_tcp_sockets, grpc_allow_network_connections,
    and grpc_disable_tls
    can possibly permit control of this computer and any data which resides on it.
    Modification of this configuration is not recommended.  Please see the
    documentation for your installed product for additional information.
    """

    def __init__(
        self,
        host: str = "127.0.0.1",
        install_path: Optional[str] = None,
        secret_key: str = "",
        grpc_port: int = 12345,
        grpc_use_tcp_sockets: Optional[bool] = False,
        grpc_allow_network_connections: Optional[bool] = False,
        grpc_disable_tls: Optional[bool] = False,
        grpc_uds_pathname: Optional[str] = None,
        html_hostname: Optional[str] = None,
        html_port: Optional[int] = None,
        ws_port: Optional[int] = None,
        session_directory: Optional[str] = None,
        timeout: float = 120.0,
        rest_api: bool = False,
        sos: bool = False,
        webui_port: Optional[int] = None,
    ) -> None:
        # every session instance needs a unique name that can be used as a cache key
        self._dsg_session: Optional["DSGSession"] = None
        self._session_name = str(uuid.uuid1())
        # when objects come into play, we can reuse them, so hash ID to instance here
        self._ensobj_hash: Dict[int, "ENSOBJ"] = {}
        self._language = "en"
        self._rest_api_enabled = rest_api
        self._sos_enabled = sos
        self._timeout = timeout
        self._cei_home = ""
        self._cei_suffix = ""
        self._hostname = host
        self._install_path = install_path
        self._launcher = None
        if html_hostname == "" or html_hostname is None:
            # if we weren't given an html host, use the hostname
            self._html_hostname = self._hostname
        else:
            self._html_hostname = html_hostname
        self._html_port = html_port
        self._ws_port = ws_port
        self._secret_key = secret_key
        self._grpc_port = grpc_port
        self._grpc_use_tcp_sockets = grpc_use_tcp_sockets
        self._grpc_allow_network_connections = grpc_allow_network_connections
        self._grpc_disable_tls = grpc_disable_tls
        self._grpc_uds_pathname = grpc_uds_pathname
        self._halt_ensight_on_close = True
        self._callbacks: Dict[str, Tuple[int, Any]] = dict()
        self._webui_port = webui_port
        # if the caller passed a session directory we will assume they are
        # creating effectively a proxy Session and create a (stub) launcher
        if session_directory is not None:
            self._launcher = Launcher()
            self._launcher.session_directory = session_directory
            # The stub will not know about us
            self._halt_ensight_on_close = False

        # are we in a jupyter notebook?
        try:
            _ = get_ipython()  # type: ignore
            self._jupyter_notebook = True  # pragma: no cover
        except NameError:
            self._jupyter_notebook = False

        # Connect to the EnSight instance
        from ansys.api.pyensight import ensight_api  # pylint: disable=import-outside-toplevel
        from ansys.pyensight.core import ensight_grpc  # pylint: disable=import-outside-toplevel

        self._ensight = ensight_api.ensight(self)
        self._build_utils_interface()
        self._grpc = ensight_grpc.EnSightGRPC(
            host=self._hostname,
            port=self._grpc_port,
            secret_key=self._secret_key,
            grpc_use_tcp_sockets=self._grpc_use_tcp_sockets,
            grpc_allow_network_connections=self._grpc_allow_network_connections,
            grpc_disable_tls=self._grpc_disable_tls,
            grpc_uds_pathname=self._grpc_uds_pathname,
            session=self,
        )
        self._grpc.session_name = self._session_name

        # establish the connection with retry
        self._establish_connection(validate=True)

        # update the enums to match current EnSight instance
        cmd = "{key: getattr(ensight.objs.enums, key) for key in dir(ensight.objs.enums)}"
        new_enums = self.cmd(cmd)
        for key, value in new_enums.items():
            if key.startswith("__") and (key != "__OBJID__"):
                continue
            setattr(self._ensight.objs.enums, key, value)

        # create ensight.core
        self._ensight.objs.core = self.cmd("ensight.objs.core")

        # get the remote Python interpreter version
        self.cmd("import platform", do_eval=False)
        self._ensight_python_version = self.cmd("platform.python_version_tuple()")

        # Because this session can have allocated significant external resources
        # we very much want a chance to close it up cleanly. It is legal to
        # call close() twice on this class if needed.
        self._already_closed = False
        atexit.register(self.close)

        # Speed up subtype lookups:
        self._subtype_tables = {}
        part_lookup_dict = dict()
        part_lookup_dict[0] = "ENS_PART_MODEL"
        part_lookup_dict[1] = "ENS_PART_CLIP"
        part_lookup_dict[2] = "ENS_PART_CONTOUR"
        part_lookup_dict[3] = "ENS_PART_DISCRETE_PARTICLE"
        part_lookup_dict[4] = "ENS_PART_FRAME"
        part_lookup_dict[5] = "ENS_PART_ISOSURFACE"
        part_lookup_dict[6] = "ENS_PART_PARTICLE_TRACE"
        part_lookup_dict[7] = "ENS_PART_PROFILE"
        part_lookup_dict[8] = "ENS_PART_VECTOR_ARROW"
        part_lookup_dict[9] = "ENS_PART_ELEVATED_SURFACE"
        part_lookup_dict[10] = "ENS_PART_DEVELOPED_SURFACE"
        part_lookup_dict[15] = "ENS_PART_BUILT_UP"
        part_lookup_dict[16] = "ENS_PART_TENSOR_GLYPH"
        part_lookup_dict[17] = "ENS_PART_FX_VORTEX_CORE"
        part_lookup_dict[18] = "ENS_PART_FX_SHOCK"
        part_lookup_dict[19] = "ENS_PART_FX_SEP_ATT"
        part_lookup_dict[20] = "ENS_PART_MAT_INTERFACE"
        part_lookup_dict[21] = "ENS_PART_POINT"
        part_lookup_dict[22] = "ENS_PART_AXISYMMETRIC"
        part_lookup_dict[24] = "ENS_PART_VOF"
        part_lookup_dict[25] = "ENS_PART_AUX_GEOM"
        part_lookup_dict[26] = "ENS_PART_FILTER"
        self._subtype_tables["ENS_PART"] = part_lookup_dict
        annot_lookup_dict = dict()
        annot_lookup_dict[0] = "ENS_ANNOT_TEXT"
        annot_lookup_dict[1] = "ENS_ANNOT_LINE"
        annot_lookup_dict[2] = "ENS_ANNOT_LOGO"
        annot_lookup_dict[3] = "ENS_ANNOT_LGND"
        annot_lookup_dict[4] = "ENS_ANNOT_MARKER"
        annot_lookup_dict[5] = "ENS_ANNOT_ARROW"
        annot_lookup_dict[6] = "ENS_ANNOT_DIAL"
        annot_lookup_dict[7] = "ENS_ANNOT_GAUGE"
        annot_lookup_dict[8] = "ENS_ANNOT_SHAPE"
        self._subtype_tables["ENS_ANNOT"] = annot_lookup_dict
        tool_lookup_dict = dict()
        tool_lookup_dict[0] = "ENS_TOOL_CURSOR"
        tool_lookup_dict[1] = "ENS_TOOL_LINE"
        tool_lookup_dict[2] = "ENS_TOOL_PLANE"
        tool_lookup_dict[3] = "ENS_TOOL_BOX"
        tool_lookup_dict[4] = "ENS_TOOL_CYLINDER"
        tool_lookup_dict[5] = "ENS_TOOL_CONE"
        tool_lookup_dict[6] = "ENS_TOOL_SPHERE"
        tool_lookup_dict[7] = "ENS_TOOL_REVOLUTION"
        self._subtype_tables["ENS_TOOL"] = tool_lookup_dict

    def __repr__(self):
        # if this is called from in the ctor, self.launcher might be None.
        session_dir = ""
        if self.launcher:
            session_dir = self.launcher.session_directory
        s = f"Session(host='{self.hostname}', secret_key='{self.secret_key}', "
        s += f"sos={self.sos}, rest_api={self.rest_api}, "
        s += f"html_hostname='{self.html_hostname}', html_port={self.html_port}, "
        s += f"grpc_port={self._grpc_port}, "
        s += f"ws_port={self.ws_port}, session_directory=r'{session_dir}', "
        s += f"webui_port={self._webui_port}, "
        s += f"grpc_use_tcp_sockets={self._grpc_use_tcp_sockets}, "
        s += f"grpc_allow_network_connections={self._grpc_allow_network_connections}, "
        s += f"grpc_disable_tls={self._grpc_disable_tls}, "
        s += f"grpc_uds_pathname='{self._grpc_uds_pathname}')"
        return s

    def _establish_connection(self, validate: bool = False) -> None:
        """Establish a gRPC connection to the EnSight instance.

        Parameters
        ----------
        validate : bool
            If true, actually try to communicate with EnSight. By default false.
        """
        time_start = time.time()
        while time.time() - time_start < self._timeout:  # pragma: no cover
            if self._grpc.is_connected():
                try:
                    if validate:
                        self._cei_home = self.cmd("ensight.version('CEI_HOME')")
                        self._cei_suffix = self.cmd("ensight.version('suffix')")
                    self._check_rest_connection()
                    return
                except OSError:  # pragma: no cover
                    pass  # pragma: no cover
            self._grpc.connect(timeout=self._timeout)
        raise RuntimeError("Unable to establish a gRPC connection to EnSight.")  # pragma: no cover

    def _check_rest_connection(self) -> None:
        """Validate the REST API connection works

        Use requests to see if the REST API is up and running (it takes time
        for websocketserver to make a gRPC connection as well).

        """
        if not self.rest_api:
            return
        #
        #
        # even when using PIM and a proxy server (Ansys Lab) this connects
        # directly from the python running in the Notebook (the front-end)
        # to the EnSight Docker Container and not the proxy server.
        # Thus, here we use 'http', the private hostname, and the html port
        # (which is the same on the proxy server).
        url = f"http://{self._hostname}:{self.html_port}/ensight/v1/session/exec"
        time_start = time.time()
        while time.time() - time_start < self._timeout:
            try:
                _ = requests.put(
                    url,
                    json="enscl.rest_test = 30*20",
                    headers=dict(Authorization=f"Bearer {self.secret_key}"),
                )
                return
            except Exception:
                pass
            time.sleep(0.5)
        raise RuntimeError("Unable to establish a REST connection to EnSight.")  # pragma: no cover

    @property
    def name(self) -> str:
        """The session name is a unique identifier for this Session instance.  It
        is used by EnSight to maintain session specific data values within the
        EnSight instance."""
        return self._session_name

    @property
    def language(self) -> str:
        """Current language specification for the EnSight session. Various
        information calls return their information in the target language
        if possible. The default is ``"en"``.

        Examples
        --------

        >>> session.language = "en"
        >>> session.ensight.objs.core.attrinfo(session.ensight.objs.enums.PREDEFINEDPALETTES)
        >>> session.language = "zh"
        >>> session.ensight.objs.core.attrinfo(session.ensight.objs.enums.PREDEFINEDPALETTES)

        """
        return self._language

    @language.setter
    def language(self, value: str) -> None:
        self._language = value
        self.cmd(f"ensight.core.tr.changelang(lin='{self._language}')", do_eval=False)

    @property
    def halt_ensight_on_close(self) -> bool:
        """Flag for indicating whether to halt EnSight on close. If this property
        is ``True`` and the session was created via a launcher, when the session
        is closed, the EnSight instance is stopped.

        .. Note::
           While this flag prevents the :func:`close<ansys.pyensight.core.Session.close>`
           method from shutting down EnSight, depending on how the host Python interpreter is configured,
           the EnSight session may still be halted. For example, this behavior can
           occur in Jupyter Lab.
        """
        return self._halt_ensight_on_close

    @halt_ensight_on_close.setter
    def halt_ensight_on_close(self, value: bool) -> None:
        self._halt_ensight_on_close = value

    @property
    def timeout(self) -> float:
        """Amount of time in seconds to try a gRPC connection before giving up."""
        return self._timeout

    @timeout.setter
    def timeout(self, value: float) -> None:
        self._timeout = value

    @property
    def cei_home(self) -> str:
        """Value of ``CEI_HOME`` for the connected EnSight session."""
        return self._cei_home

    @property
    def cei_suffix(self) -> str:
        """Suffix string of the connected EnSight session. For example, ``222``."""
        return self._cei_suffix

    @property
    def jupyter_notebook(self) -> bool:
        """Flag indicating if the session is running in a Jupyter notebook and should use
        the display features of that interface.

        """
        return self._jupyter_notebook

    @jupyter_notebook.setter
    def jupyter_notebook(self, value: bool) -> None:
        self._jupyter_notebook = value

    @property
    def ensight(self) -> "ensight_api.ensight":
        """Core EnSight API wrapper."""
        return self._ensight

    @property
    def grpc(self) -> "ensight_grpc.EnSightGRPC":
        """The gRPC wrapper instance used by this session to access EnSight."""
        return self._grpc

    @property
    def secret_key(self) -> str:
        """Secret key used for communication validation in the gRPC instance."""
        return self._secret_key

    @property
    def html_port(self) -> Optional[int]:
        """Port supporting HTML interaction with EnSight."""
        return self._html_port

    @property
    def ws_port(self) -> Optional[int]:
        """Port supporting WS interaction with EnSight."""
        return self._ws_port

    @property
    def hostname(self) -> str:
        """Hostname of the system hosting the EnSight instance."""
        return self._hostname

    @property
    def html_hostname(self) -> str:
        """Hostname of the system hosting the EnSight web server instance."""
        return self._html_hostname

    @property
    def launcher(self) -> "Launcher":
        """Reference to the launcher instance if a launcher was used to instantiate the session."""
        return self._launcher

    @launcher.setter
    def launcher(self, value: "Launcher"):
        self._launcher = value

    @property
    def sos(self) -> bool:
        """
        Flag indicating if the remote EnSight session is running in SOS (Server of Server) mode.
        """
        return self._sos_enabled

    @property
    def rest_api(self) -> bool:
        """
        Flag indicating if the remote EnSight session supports the REST API.
        """
        return self._rest_api_enabled

    @staticmethod
    def help():
        """Open the documentation for PyEnSight in a web browser."""
        url = "https://ensight.docs.pyansys.com/"
        webbrowser.open(url)

    def copy_to_session(
        self,
        local_prefix: str,
        filelist: List[str],
        remote_prefix: Optional[str] = None,
        progress: bool = False,
    ) -> list:
        """Copy a collection of files into the EnSight session.

        Copy files from the local filesystem into the filesystem that is hosting
        the EnSight instance.

        .. note::
           For a :class:`LocalLauncheransys.pyensight.core.LocalLauncher>`
           instance, these are the same filesystems.

        Parameters
        ----------
        local_prefix : str
            URL prefix to use for all files specified for the ``filelist``
            parameter. The only protocol supported is ``'file://'``, which
            is the local filesystem.
        filelist : list
            List of files to copy. These files are prefixed with ``local_prefix``
            and written relative to the ``remote_prefix`` parameter appended to
            ``session.launcher.session_directory``.
        remote_prefix : str
            Directory on the remote (EnSight) filesystem, which is the
            destination for the files. This prefix is appended to
            ``session.launcher.session_directory``.
        progress : bool, optional
            Whether to show a progress bar. The default is ``False``. If ``True`` and
            the ``tqdm`` module is available, a progress bar is shown.

        Returns
        -------
        list
            List of the filenames that were copied and their sizes.

        Examples
        --------
        >>> the_files = ["fluent_data_dir", "ensight_script.py"]
        >>> session.copy_to_session("file:///D:/data", the_files, progress=True)

        >>> the_files = ["fluent_data_dir", "ensight_script.py"]
        >>> session.copy_to_session("file:///scratch/data", the_files, remote_prefix="data")

        """
        uri = urlparse(local_prefix)
        if uri.scheme != "file":
            raise RuntimeError("Only the file:// protocol is supported for the local_prefix")
        localdir = url2pathname(uri.path)

        remote_functions = textwrap.dedent(
            """\
                import os
                def copy_write_function__(filename: str, data: bytes) -> None:
                    os.makedirs(os.path.dirname(filename), exist_ok=True)
                    with open(filename, "ab") as fp:
                        fp.write(data)
            """
        )

        self.cmd(remote_functions, do_eval=False)

        out = []
        dirlen = 0
        if localdir:  # pragma: no cover
            # we use dirlen + 1 here to remove the '/' inserted by os.path.join()
            dirlen = len(localdir) + 1
        for item in filelist:
            try:
                name = os.path.join(localdir, item)
                if os.path.isfile(name):
                    out.append((name[dirlen:], os.stat(name).st_size))
                else:
                    for root, _, files in os.walk(name):
                        for filename in files:
                            fullname = os.path.join(root, filename)
                            out.append((fullname[dirlen:], os.stat(fullname).st_size))
            except Exception:
                pass
        if progress:  # pragma: no cover
            try:
                from tqdm.auto import tqdm
            except ImportError:
                tqdm = list
        else:
            tqdm = list
        for item in tqdm(out):
            filename = os.path.join(localdir, item[0])
            out_dir = self.launcher.session_directory.replace("\\", "/")
            if remote_prefix:
                out_dir += f"/{remote_prefix}"
            name = out_dir + f"/{item[0]}"
            name = name.replace("\\", "/")
            # Walk the file in chunk size blocks
            chunk_size = 1024 * 1024
            with open(filename, "rb") as fp:
                while True:
                    data = fp.read(chunk_size)
                    if data == b"":
                        break
                    self.cmd(
                        f"copy_write_function__(r'{name}', {data!r})", do_eval=False
                    )  # pragma: no cover
        return out

    def copy_from_session(
        self,
        local_prefix: str,
        filelist: List[str],
        remote_prefix: Optional[str] = None,
        progress: bool = False,
    ) -> list:
        """Copy a collection of files out of the EnSight session.

        Copy files from the filesystem of the remote EnSight instance to the
        filesystem of the local PyEnsight instance.

        .. note::
           For a :class:`LocalLauncheransys.pyensight.core.LocalLauncher>`
           instance, these are the same filesystems.

        Parameters
        ----------
        local_prefix : str
            URL prefix of the location to save the files to. The only
            protocol currently supported is ``'file://'``, which is the
            local filesystem.
        filelist : list
            List of the files to copy. These files are prefixed
                with ``session.launcher.session_directory/remote_prefix`` and written
                relative to URL prefix specified for the ``local_prefix`` parameter.
        remote_prefix : str, optional
            Directory on the remote (EnSight) filesystem that is the source
            for the files. This prefix is appended to ``session.launcher.session_directory``.
        progress : bool, optional
            Whether to show a progress bar. The default is ``False``. If ``True`` and
            the ``tqdm`` module is available, a progress bar is shown.

        Returns
        -------
        list
            List of the files that were copied.

        Examples
        --------
        >>> the_files = ["fluent_data_dir", "ensight_script.py"]
        >>> session.copy_from_session("file:///D:/restored_data", the_files, progress=True)

        >>> the_files = ["fluent_data_dir", "ensight_script.py"]
        >>> session.copy_from_session("file:///scratch/restored_data", the_files,
                remote_prefix="data")
        """

        uri = urlparse(local_prefix)
        if uri.scheme != "file":
            raise RuntimeError("Only the file:// protocol is supported for the local_prefix")
        localdir = url2pathname(uri.path)

        remote_functions = textwrap.dedent(
            """\
                import os
                def copy_walk_function__(remotedir: str, filelist: list) -> None:
                    out = []
                    dirlen = 0
                    if remotedir:
                        dirlen = len(remotedir) + 1
                    for item in filelist:
                        try:
                            name = os.path.join(remotedir, item)
                            if os.path.isfile(name):
                                out.append((name[dirlen:], os.stat(name).st_size))
                            else:
                                for root, _, files in os.walk(name):
                                    for filename in files:
                                        fullname = os.path.join(root, filename)
                                        out.append((fullname[dirlen:], os.stat(fullname).st_size))
                        except Exception:
                            pass
                    return out
                # (needed for flake8)
                def copy_read_function__(filename: str, offset: int, numbytes: int) -> bytes:
                    with open(filename, "rb") as fp:
                        fp.seek(offset)
                        data = fp.read(numbytes)
                    return data
            """
        )

        self.cmd(remote_functions, do_eval=False)

        remote_directory = self.launcher.session_directory
        if remote_prefix:
            remote_directory = f"{remote_directory}/{remote_prefix}"
        remote_directory = remote_directory.replace("\\", "/")
        names = self.cmd(f"copy_walk_function__(r'{remote_directory}', {filelist})", do_eval=True)
        if progress:
            try:
                from tqdm.auto import tqdm
            except ImportError:
                tqdm = list
        else:
            tqdm = list
        for item in tqdm(names):
            name = f"{remote_directory}/{item[0]}".replace("\\", "/")
            full_name = os.path.join(localdir, item[0])
            os.makedirs(os.path.dirname(full_name), exist_ok=True)
            with open(full_name, "wb") as fp:
                offset = 0
                chunk_size = 1024 * 1024
                while True:
                    data = self.cmd(
                        f"copy_read_function__(r'{name}', {offset}, {chunk_size})", do_eval=True
                    )
                    if len(data) == 0:
                        break
                    fp.write(data)
                    offset += chunk_size
        return names

    def run_script(self, filename: str) -> Optional[types.ModuleType]:
        """Run an EnSight Python script file.

        In EnSight, there is a notion of a Python *script* that is normally run line by
        line in EnSight. In such scripts, the ``ensight`` module is assumed to be preloaded.
        This method runs such scripts by importing them as modules and running the commands
        through the PyEnSight interface. This is done by installing the PyEnsight ``Session``
        object into the module before it is imported. This makes it possible to use a
        Python debugger with an EnSight Python script, using the PyEnSight interface.

        .. note::

            Because the Python script is imported as a module, the script filename must
            have a ``.py`` extension.


        Parameters
        ----------
        filename : str
            Filename of the Python script to run, which is loaded as a module by PyEnSight.

        Returns
        -------
        types.ModuleType
            Imported module.

        """
        dirname = os.path.dirname(filename)
        if not dirname:  # pragma: no cover
            dirname = "."  # pragma: no cover
        if dirname not in sys.path:
            sys.path.append(dirname)
        module_name, _ = os.path.splitext(os.path.basename(filename))
        # get the module reference
        spec = importlib.util.find_spec(module_name)
        if spec:  # pragma: no cover
            module = importlib.util.module_from_spec(spec)
            # insert an ensight interface into the module
            if self.ensight:
                module.ensight = self.ensight  # type: ignore
                # load (run) the module
                if spec.loader:  # pragma: no cover
                    spec.loader.exec_module(module)
            return module
        return None  # pragma: no cover

    def exec(self, function: Callable, *args, remote: bool = False, **kwargs) -> Any:
        """Run a function containing EnSight API calls locally or in the EnSight interpreter.

        The function is in this form::

            def myfunc(ensight, *args, **kwargs):
                ...
                return value

        The ``exec()`` method allows for the function to be executed in the PyEnSight Python
        interpreter or the (remote) EnSight interpreter. Thus, a function making a large
        number of RPC calls can run much faster than if it runs solely in the PyEnSight
        interpreter.

        These constraints exist on this capability:

        - The function may only use arguments passed to the ``exec()`` method and can only
          return a single value.
        - The function cannot modify the input arguments.
        - The input arguments must be serializable and the PyEnSight Python interpreter
          version must match the version in EnSight.

        Parameters
        ----------
        remote : bool, optional
            Whether to execute the function in the (remote) EnSight interpreter.

        Examples
        --------
        >>> from ansys.pyensight.core import LocalLauncher
        >>> session = LocalLauncher().start()
        >>> options = dict()
        >>> options['Verbose mode'] = 'OFF'
        >>> options['Use ghost elements'] = 'OFF'
        >>> options['Long names'] = 'OFF'
        >>> options['Compatibility mode'] = 'ON'
        >>> options['Move Transient Parts'] = 'ON'
        >>> options['Element type'] = 'Tri 3'
        >>> options['Boundary ghosts'] = 'None'
        >>> options['Spread out parts'] = 'Legacy'
        >>> options['Number of spheres'] = 100
        >>> options['Number of cubes'] = 100
        >>> options['Number of planes'] = 0
        >>> options['Number of elements start'] = 1000
        >>> options['Number of elements end'] = 1000
        >>> options['Number of timesteps'] = 1
        >>> options['Part scaling factor'] = 1.000000e+00
        >>> options['Random number seed'] = 0
        >>> options['Number of scalars'] = 3
        >>> options['Number of vectors'] = 3
        >>> options['Number of constants'] = 3
        >>> session.load_data("dummy", file_format="Synthetic", reader_options=options)

        >>> def count(ensight, attr, value):
        >>>     import time
        >>>     start = time.time()
        >>>     count = 0
        >>>     for p in ensight.objs.core.PARTS:
        >>>         if p.getattr(attr) == value:
        >>>             count += 1
        >>> print(count(session.ensight, "VISIBLE", True))
        >>> print(session.exec(count, "VISIBLE", True))
        >>> print(session.exec(count, "VISIBLE", True, remote=True))

        """
        if remote:
            # remote execution only supported in 2023 R1 or later
            if int(self._cei_suffix) < 231:
                raise RuntimeError("Remote function execution only supported in 2023 R1 and later")
            local_python_version = platform.python_version_tuple()
            if self._ensight_python_version[0:2] != local_python_version[0:2]:
                vers = "Local and remote Python versions must match: "
                vers += ".".join(local_python_version)
                vers += " vs "
                vers += ".".join(self._ensight_python_version)
                raise RuntimeError(vers)
            import dill  # pylint: disable=import-outside-toplevel

            # Create a bound object that allows for direct encoding of the args/kwargs params
            # The new function would be bound_function(ensight) where the args are captured
            # in the lambda.
            bound_function = lambda ens: function(  # noqa: E731  # pragma: no cover
                ens, *args, **kwargs
            )
            # Serialize the bound function
            serialized_function = dill.dumps(bound_function, recurse=True)
            self.cmd("import dill", do_eval=False)
            # Run it remotely, passing the instance ensight instead of self._ensight
            cmd = f"dill.loads(eval(repr({serialized_function})))(ensight)"
            return self.cmd(cmd)
        else:
            return function(self._ensight, *args, **kwargs)

    def show(
        self,
        what: str = "image",
        width: Optional[int] = None,
        height: Optional[int] = None,
        temporal: bool = False,
        aa: int = 4,
        fps: float = 30.0,
        num_frames: Optional[int] = None,
        ui: Optional[str] = "simple",
    ) -> "renderable.Renderable":
        """Capture the current EnSight scene or otherwise make it available for
        display in a web browser.

        This method generates the appropriate visuals and returns the renderable
        object for viewing. If the session is in a Jupyter notebook, the cell
        in which the ``show()`` method is issued is updated with the renderable display.

        Parameters
        ----------
        what : str, optional
            Type of scene display to generate. The default is ``"image"``.
            Options are:

            * ``image``: Simple rendered PNG image
            * ``deep_pixel``: EnSight deep pixel image
            * ``animation``: MPEG4 movie
            * ``webgl``: Interactive WebGL-based browser viewer
            * ``remote``: Remote rendering-based interactive EnSight viewer
            * ``remote_scene``: Remote rendering-based interactive EnSight viewer

        width : int, optional
            Width of the rendered entity. The default is ``None``.
        height : int, optional
            Height of the rendered entity. The default is ``None``.
        temporal : bool, optional
            Whether to include all timesteps in WebGL views. The default is ``False``.
        aa : int, optional
            Number of antialiasing passes to use when rendering images. The
            default is ``4``.
        fps : float, optional
            Number of frames per second to use for animation playback. The default
            is ``30``.
        num_frames : int, optional
            Number of frames of static timestep to record for animation playback.

        Returns
        -------
        renderable.Renderable

        Raises
        ------
        RuntimeError
            If it is not possible to generate the content.

        Examples
        --------
        Render an image and display it in a browser. Rotate the scene and update the display.

        >>> image = session.show('image', width=800, height=600)
        >>> image.browser()
        >>> session.ensight.view_transf.rotate(30, 30, 0)
        >>> image.update()
        >>> image.browser()

        """
        self._establish_connection()
        if self._html_port is None:
            raise RuntimeError("No websocketserver has been associated with this Session")

        kwargs: Dict[str, Union[float, int, None, str]] = dict(
            height=height, width=width, temporal=temporal, aa=aa, fps=fps, num_frames=num_frames
        )
        if self._jupyter_notebook:  # pragma: no cover
            from IPython.display import display

            # get the cell DisplayHandle instance
            kwargs["cell_handle"] = display("", display_id=True)

        render = None
        if what == "image":
            render = RenderableImage(self, **kwargs)
        elif what == "deep_pixel":
            render = RenderableDeepPixel(self, **kwargs)
        elif what == "animation":
            render = RenderableMP4(self, **kwargs)
        elif what == "webgl":
            render = RenderableWebGL(self, **kwargs)
        elif what == "sgeo":
            # the SGEO protocol is only supported in 2023 R1 and higher
            if int(self._cei_suffix) < 231:
                # Use the AVZ viewer in older versions of EnSight
                render = RenderableWebGL(self, **kwargs)
            else:
                render = RenderableSGEO(self, **kwargs)
        elif what == "remote":
            kwargs["ui"] = ui
            render = RenderableVNC(self, **kwargs)
        elif what == "remote_scene":
            render = RenderableEVSN(self, **kwargs)
        # Undocumented. Available only internally
        elif what == "webensight":
            render = RenderableVNCAngular(self, **kwargs)
        elif what == "webui":
            render = RenderableFluidsWebUI(self, **kwargs)

        if render is None:
            raise RuntimeError("Unable to generate requested visualization")

        return render

    def cmd(self, value: str, do_eval: bool = True) -> Any:
        """Run a command in EnSight and return the results.

        Parameters
        ----------
        value : str
            String of the command to run
        do_eval : bool, optional
            Whether to perform an evaluation. The default is ``True``.


        Returns
        -------
        result
            Result of the string being executed as Python inside EnSight.

        Examples
        --------

        >>> print(session.cmd("10+4"))
            14
        """
        if self._dsg_session:
            self._dsg_session._pyensight_grpc_coming = True
        self._establish_connection()
        ret = self._grpc.command(value, do_eval=do_eval)
        if self._dsg_session:
            self._dsg_session._pyensight_grpc_coming = False
        if do_eval:
            ret = self._convert_ctor(ret)
            value = eval(ret, dict(session=self, ensobjlist=ensobjlist))
            return value
        return ret

    def geometry(self, what: str = "glb") -> bytes:
        """Return the current EnSight scene as a geometry file.

        Parameters
        ----------
        what : str, optional
            File format to return. The default is ``"glb"``.

        Returns
        -------
        obj
            Generated geometry file as a bytes object.

        Examples
        --------
        >>> data = session.geometry()
        >>> with open("file.glb", "wb") as fp:
        >>> fp.write(data)

        """
        self._establish_connection()
        return self._grpc.geometry()

    def render(self, width: int, height: int, aa: int = 1) -> bytes:
        """Render the current EnSight scene and return a PNG image.

        Parameters
        ----------
        width : int
            Width of the rendered image in pixels.
        height : int
            Height of the rendered image in pixels.
        aa : int, optional
            Number of antialiasing passes to use. The default is ``1``.

        Returns
        -------
        obj
            PNG image as a bytes object.

        Examples
        --------
        >>> data = session.render(1920, 1080, aa=4)
        >>> with open("file.png", "wb") as fp:
        >>> fp.write(data)

        """
        self._establish_connection()
        return self._grpc.render(width=width, height=height, aa=aa)

    def _release_remote_objects(self, object_id: Optional[int] = None):
        """
        Send a command to the remote EnSight session to drop a specific object
        or all objects from the remote object cache.

        Parameters
        ----------
        object_id: int, optional
            The specific object to drop from the cache.  If no objects are specified,
            then all remote objects associated with this session will be dropped.

        """
        obj_str = ""
        if object_id:  # pragma: no cover
            obj_str = f", id={object_id}"  # pragma: no cover
        cmd = f"ensight.objs.release_id('{self.name}'{obj_str})"
        _ = self.cmd(cmd, do_eval=False)

    def close(self) -> None:
        """Close the session.

        Close the current session and its gRPC connection.
        """
        # if version 242 or higher, free any objects we have cached there
        if not self._already_closed:
            if self.cei_suffix >= "242":
                try:
                    self._release_remote_objects()
                except RuntimeError:  # pragma: no cover
                    # handle some intermediate EnSight builds.
                    pass
                except IOError:  # pragma: no cover
                    # The session might already have been closed via another
                    # session object. If grpc is inactive, there's no sense
                    # in raising an exception since we are closing it anyway
                    pass
            if self._launcher and self._halt_ensight_on_close:
                self._launcher.close(self)
            else:
                # lightweight shtudown, just close the gRC connection
                self._grpc.shutdown(stop_ensight=False)
            self._already_closed = True
        self._launcher = None

    def _build_utils_interface(self) -> None:
        """Build the ``ensight.utils`` interface.

        This method Walk the PY files in the ``utils`` directory, creating instances
        of the classes in those files and placing them in the
        ``Session.ensight.utils`` namespace.
        """
        self._ensight.utils = types.SimpleNamespace()
        _utils_dir = os.path.join(os.path.dirname(__file__), "utils")
        if _utils_dir not in sys.path:
            sys.path.insert(0, _utils_dir)
        onlyfiles = [f for f in listdir(_utils_dir) if os.path.isfile(os.path.join(_utils_dir, f))]
        for _basename in onlyfiles:
            # skip over any files with the "_server" in their names
            if "_server" in _basename or "_cli" in _basename:
                continue
            _filename = os.path.join(_utils_dir, _basename)
            try:
                # get the module and class names
                _name = os.path.splitext(os.path.basename(_filename))[0]
                if _name == "__init__":
                    continue
                _cap_name = _name[0].upper() + _name[1:]
                # import the module
                spec = importlib.util.spec_from_file_location(
                    f"ansys.pyensight.core.utils.{_name}", _filename
                )
                if spec:  # pragma: no cover
                    _module = importlib.util.module_from_spec(spec)
                    if spec.loader:  # pragma: no cover
                        spec.loader.exec_module(_module)
                    # get the class from the module (query.py filename -> Query() object)
                    _the_class = getattr(_module, _cap_name)
                    # Create an instance, using ensight as the EnSight interface
                    # and place it in this module.
                    setattr(self._ensight.utils, _name, _the_class(self._ensight))
            except Exception as e:  # pragma: no cover
                # Warn on import errors
                print(f"Error loading ensight.utils from: '{_filename}' : {e}")

    MONITOR_NEW_TIMESTEPS_OFF = "off"
    MONITOR_NEW_TIMESTEPS_STAY_AT_CURRENT = "stay_at_current"
    MONITOR_NEW_TIMESTEPS_JUMP_TO_END = "jump_to_end"

    def load_data(
        self,
        data_file: str,
        result_file: Optional[str] = None,
        file_format: Optional[str] = None,
        reader_options: Optional[dict] = None,
        new_case: bool = False,
        representation: str = "3D_feature_2D_full",
        monitor_new_timesteps: str = MONITOR_NEW_TIMESTEPS_OFF,
    ) -> None:
        """Load a dataset into the EnSight instance.

        Load the data from a given file into EnSight. The new data
        replaces any currently loaded data in the session.

        Parameters
        ----------
        data_file : str
            Name of the data file to load.
        result_file : str, optional
            Name of the second data file for dual-file datasets.
        file_format : str, optional
            Name of the EnSight reader to use. The default is ``None``,
            in which case EnSight selects a reader.
        reader_options : dict, optional
            Dictionary of reader-specific option-value pairs that can be used
            to customize the reader behavior. The default is ``None``.
        new_case : bool, optional
            Whether to load the dataset in another case. The default is ``False``,
            in which case the dataset replaces the one (if any) loaded in the existing
            current case.
        representation : str, optional
            Default representation for the parts loaded. The default is
            ``"3D_feature_2D_full"``.
        monitor_new_timesteps: str, optional
            Defaulted to off, if changed EnSight will monitor for new timesteps.
            The allowed values are MONITOR_NEW_TIMESTEPS_OFF, MONITOR_NEW_TIMESTEPS_STAY_AT_CURRENT
            and MONITOR_NEW_TIMESTEPS_JUMP_TO_END

        Raises
        ------
        RuntimeError
            If EnSight cannot guess the file format or an error occurs while the
            data is being read.

        Examples
        --------
        >>> from ansys.pyensight.core import LocalLauncher
        >>> session = LocalLauncher().start()
        >>> session.load_data('D:\\data\\CFX\\example_data.res')

        """
        self._establish_connection()
        # what application are we talking to?
        target = self.cmd("ensight.version('product').lower()")
        if target == "envision":
            cmd = f'ensight.data.replace(r"""{data_file}""")'
            if self.cmd(cmd) != 0:
                raise RuntimeError("Unable to load the dataset.")
            return

        # Handle case changes...
        cmds = [
            'ensight.case.link_modelparts_byname("OFF")',
            'ensight.case.create_viewport("OFF")',
            'ensight.case.apply_context("OFF")',
            "ensight.case.reflect_model_in(\"'none'\")",
        ]
        for cmd in cmds:
            self.cmd(cmd, do_eval=False)

        if new_case:
            # New case
            new_case_name = None
            for case in self.ensight.objs.core.CASES:
                if case.ACTIVE == 0:
                    new_case_name = case.DESCRIPTION
                    break
            if new_case_name is None:
                raise RuntimeError("No cases available for adding.")
            cmd = f'ensight.case.add("{new_case_name}")'
            self.cmd(cmd, do_eval=False)
            cmd = f'ensight.case.select("{new_case_name}")'
            self.cmd(cmd, do_eval=False)
        else:
            # Case replace
            current_case_name = self.ensight.objs.core.CURRENTCASE[0].DESCRIPTION
            cmd = f'ensight.case.replace("{current_case_name}", "{current_case_name}")'
            self.cmd(cmd, do_eval=False)
            cmd = f'ensight.case.select("{current_case_name}")'
            self.cmd(cmd, do_eval=False)

        # Attempt to find the file format if none is specified
        if file_format is None:
            try:
                cmd = "ensight.objs.core.CURRENTCASE[0]"
                cmd += f'.queryfileformat(r"""{data_file}""")["reader"]'
                file_format = self.cmd(cmd)
            except RuntimeError:
                raise RuntimeError(f"Unable to determine file format for {data_file}")

        # Load the data
        cmds = [
            "ensight.part.select_default()",
            "ensight.part.modify_begin()",
            f'ensight.part.elt_representation("{representation}")',
            "ensight.part.modify_end()",
            'ensight.data.binary_files_are("native")',
            f'ensight.data.format("{file_format}")',
        ]
        if reader_options:
            for key, value in reader_options.items():
                option = f"""ensight.data.reader_option("{repr(key)} {repr(value)}")"""
                cmds.append(option)
        if result_file:
            cmds.append(f'ensight.data.result(r"""{result_file}""")')
        cmds.append("ensight.data.shift_time(1.000000, 0.000000, 0.000000)")
        cmds.append(f'ensight.solution_time.monitor_for_new_steps("{monitor_new_timesteps}")')
        cmds.append(f'ensight.data.replace(r"""{data_file}""")')
        for cmd in cmds:
            if self.cmd(cmd) != 0:
                raise RuntimeError("Unable to load the dataset.")

    def download_pyansys_example(
        self,
        filename: str,
        directory: Optional[str] = None,
        root: Optional[str] = None,
        folder: Optional[bool] = None,
    ) -> str:
        """Download an example dataset from the ansys/example-data repository.
        The dataset is downloaded local to the EnSight server location, so that it can
        be downloaded even if running from a container.

        Parameters
        ----------
        filename: str
            The filename to download
        directory: str
            The directory to download the filename from
        root: str
            If set, the download will happen from another location
        folder: bool
            If set to True, it marks the filename to be a directory rather
            than a single file

        Returns
        -------
        pathname: str
            The download location, local to the EnSight server directory.
            If folder is set to True, the download location will be a folder containing
            all the items available in the repository location under that folder.

        Examples
        --------
        >>> from ansys.pyensight.core import DockerLauncher
        >>> session = DockerLauncher().start(data_directory="D:\\")
        >>> cas_file = session.download_pyansys_example("mixing_elbow.cas.h5","pyfluent/mixing_elbow")
        >>> dat_file = session.download_pyansys_example("mixing_elbow.dat.h5","pyfluent/mixing_elbow")
        >>> session.load_data(cas_file, result_file=dat_file)
        >>> remote = session.show("remote")
        >>> remote.browser()
        """
        base_uri = "https://api.github.com/repos/ansys/example-data/contents"
        override_root = False
        if not folder:
            if root is not None:
                base_uri = root
                override_root = True
        uri = f"{base_uri}/{filename}"
        if directory:
            uri = f"{base_uri}/{directory}/{filename}"
        pathname = f"{self.launcher.session_directory}/{filename}"
        if not folder:
            if override_root:
                correct_url = uri
            else:
                correct_url = None
                with requests.get(uri) as r:
                    data = r.json()
                    correct_url = data["download_url"]
                if not correct_url:
                    raise RuntimeError(f"Couldn't retrieve download URL from github uri {uri}")
            script = "import requests\n"
            script += "import shutil\n"
            script += "import os\n"
            script += f'url = "{correct_url}"\n'
            script += f'outpath = r"""{pathname}"""\n'
            script += "with requests.get(url, stream=True) as r:\n"
            script += "    with open(outpath, 'wb') as f:\n"
            script += "        shutil.copyfileobj(r.raw, f)\n"
            self.cmd(script, do_eval=False)
        else:
            script = "import requests\n"
            script += "import shutil\n"
            script += "import os\n"
            script += f'url = "{uri}"\n'
            script += "with requests.get(url) as r:\n"
            script += "    data = r.json()\n"
            script += f'    output_directory = r"""{pathname}"""\n'
            script += "    os.makedirs(output_directory, exist_ok=True)\n"
            script += "    for item in data:\n"
            script += "        if item['type'] == 'file':\n"
            script += "            file_url = item['download_url']\n"
            script += "            filename = os.path.join(output_directory, item['name'])\n"
            script += "            r = requests.get(file_url, stream=True)\n"
            script += "            with open(filename, 'wb') as f:\n"
            script += "                f.write(r.content)\n"
            self.cmd(script, do_eval=False)
        return pathname

    def load_example(
        self, example_name: str, uncompress: bool = False, root: Optional[str] = None
    ) -> str:
        """Load an example dataset.

        This method downloads an EnSight session file from a known location and loads
        it into the current EnSight instance. The URL for the dataset is formed by
        combining the value given for the ``example_name`` parameter with a root URL.
        The default base URL is provided by Ansys, but it can be overridden by specifying
        a value for the ``root`` parameter.

        Parameters
        ----------
        example_name : str
            Name of the EnSight session file (``.ens``) to download and load.
        uncompress : bool, optional
            Whether to unzip the downloaded file into the returned directory name.
            The default is ``False``.
        root : str, optional
            Base URL for the download.

        Returns
        -------
        str
            Path to the downloaded file in the EnSight session.

        Examples
        --------
        >>> from ansys.pyensight.core import LocalLauncher
        >>> session = LocalLauncher().start()
        >>> session.load_example("fluent_wing_example.ens")
        >>> remote = session.show("remote")
        >>> remote.browser()

        """
        base_uri = "https://s3.amazonaws.com/www3.ensight.com/PyEnSight/ExampleData"
        if root is not None:  # pragma: no cover
            base_uri = root  # pragma: no cover
        pathname = self.download_pyansys_example(example_name, root=base_uri)
        script = f'outpath = r"""{pathname}"""\n'
        if uncompress:
            # in this case, remove the extension and unzip the file
            pathname_dir = os.path.splitext(pathname)[0]
            script += "outpath_dir = os.path.splitext(outpath)[0]\n"
            script += "os.mkdir(outpath_dir)\n"
            script += "shutil.unpack_archive(outpath, outpath_dir, 'zip')\n"
            # return the directory name
            pathname = pathname_dir
        else:
            script += "ensight.objs.ensxml_restore_file(outpath)\n"
        self.cmd(script, do_eval=False)
        return pathname

    def add_callback(
        self, target: Any, tag: str, attr_list: list, method: Callable, compress: bool = True
    ) -> None:
        """Register a callback with an event tuple.

        For a given target object (such as ``"ensight.objs.core"``) and a list
        of attributes (such as ``["PARTS", "VARIABLES"]``), this method sets up a
        callback to be made when any of those attribute change on the target object.
        The target can also be an EnSight (not PyEnSight) class name, for example
        "ENS_PART".  In this latter form, all objects of that type are watched for
        specified attribute changes.

        The callback is made with a single argument, a string encoded in URL format
        with the supplied tag, the name of the attribute that changed and the UID
        of the object that changed.  The string passed to the callback is in this form:
        ``grpc://{sessionguid}/{tag}?enum={attribute}&uid={objectid}``.

        Only one callback with the noted tag can be used in the session.

        Parameters
        ----------
        target : obj, str
            Name of the target object or name of a class as a string to
            match all objects of that class. A proxy class reference is
            also allowed. For example, ``session.ensight.objs.core``.
        tag : str
            Unique name for the callback. A tag can end with macros of
            the form ``{{attrname}}`` to return the value of an attribute of the
            target object. The macros should take the form of URI queries to
            simplify parsing.
        attr_list : list
            List of attributes of the target that are to result in the callback
            being called if changed.
        method : Callable
            Callable that is called with the returned URL.
        compress : bool, optional
            Whether to call only the last event if a repeated event is generated
            as a result of an action. The default is ``True``. If ``False``, every
            event results in a callback.

        Examples
        --------
        A string similar to this is printed when the dataset is loaded and the part list
        changes:

        ``    Event : grpc://f6f74dae-f0ed-11ec-aa58-381428170733/partlist?enum=PARTS&uid=221``

        >>> from ansys.pyensight.core import LocalLauncher
        >>> s = LocalLauncher().start()
        >>> def cb(v: str):
        >>>     print("Event:", v)
        >>> s.add_callback("ensight.objs.core", "partlist", ["PARTS"], cb)
        >>> s.load_data("D:\\ANSYSDev\\data\\CFX\\HeatingCoil_001.res")
        """
        self._establish_connection()
        # shorten the tag up to the query block. Macros are only legal in the query block.
        try:
            idx = tag.index("?")
            short_tag = tag[:idx]
        except ValueError:
            short_tag = tag
        if short_tag in self._callbacks:
            raise RuntimeError(f"A callback for tag '{short_tag}' already exists")
        # Build the addcallback string against the full tag
        flags = ""
        if compress:
            flags = ",flags=ensight.objs.EVENTMAP_FLAG_COMP_GLOBAL"
        if hasattr(target, "__OBJID__"):
            target = self.remote_obj(target.__OBJID__)
        cmd = f"ensight.objs.addcallback({target},None,"
        cmd += f"'{self._grpc.prefix()}{tag}',attrs={repr(attr_list)}{flags})"
        callback_id = self.cmd(cmd)
        # if this is the first callback, start the event stream
        if len(self._callbacks) == 0:
            self._grpc.event_stream_enable(callback=self._event_callback)
        # record the callback id along with the callback
        # if the callback URL starts with the short_tag, we make the callback
        self._callbacks[short_tag] = (callback_id, method)

    def remove_callback(self, tag: str) -> None:
        """Remove a callback that the :func`add_callback<ansys.pyensight.core.Session.add_callback>`
        method started.

        Given a tag used to register a previous callback (``add_callback()``), remove
        this callback from the EnSight callback system.

        Parameters
        ----------
        tag : str
            Callback string tag.

        Raises
        ------
        RuntimeError
            If an invalid tag is supplied.

        """
        if tag not in self._callbacks:
            raise RuntimeError(f"A callback for tag '{tag}' does not exist")
        callback_id = self._callbacks[tag][0]
        del self._callbacks[tag]
        cmd = f"ensight.objs.removecallback({callback_id})"
        _ = self.cmd(cmd, do_eval=False)

    def _event_callback(self, cmd: str) -> None:
        """Pass the URL back to the registered callback.

        This method matches the ``cmd`` URL with the registered callback and then
        makes the callback.

        Parameters
        ----------
        cmd : str
            URL callback from the gRPC event stream. The URL has this
            form: ``grpc://{sessionguid}/{tag}?enum={attribute}&uid={objectid}``.

        """
        # EnSight will always tack on '?enum='.  If our tag uses ?macro={{attr}},
        # you will get too many '?' in the URL, making it difficult to parse.
        # So, we look for "?..." and a following "?enum=".  If we see this, convert
        # "?enum=" into "&enum=".
        idx_question = cmd.find("?")
        idx_enum = cmd.find("?enum=")
        if idx_question < idx_enum:
            cmd = cmd.replace("?enum=", "&enum=")
        parse = urlparse(cmd)
        tag = parse.path[1:]
        for key, value in self._callbacks.items():
            # remember "key" is a shortened version of tag
            if tag.startswith(key):
                value[1](cmd)
                return
        print(f"Unhandled event: {cmd}")

    # Object API helper functions
    @staticmethod
    def remote_obj(ensobjid: int) -> str:
        """Generate a string that, for a given ``ENSOBJ`` object ID, returns
        a proxy object instance.

        Parameters
        ----------
        ensobjid: int
            ID of the ``ENSOBJ`` object.

        Returns
        -------
        str
            String for the proxy object instance.
        """
        return f"ensight.objs.wrap_id({ensobjid})"

    def _prune_hash(self) -> None:
        """Prune the ``ENSOBJ`` hash table.

        The ``ENSOBJ`` hash table may need flushing if it gets too big. Do that here."""
        if len(self._ensobj_hash) > 1000000:
            self._ensobj_hash = {}

    def add_ensobj_instance(self, obj: "ENSOBJ") -> None:
        """Add a new ``ENSOBJ`` object instance to the hash table.

        Parameters
        ----------
        obj : ENSOBJ
           ``ENSOBJ`` object instance.
        """
        self._ensobj_hash[obj.__OBJID__] = obj

    def obj_instance(self, ensobjid: int) -> Optional["ENSOBJ"]:
        """Get any existing proxy object associated with an ID.

        Parameters
        ----------
        ensobjid: int
            ID of the ``ENSOBJ`` object.

        """
        return self._ensobj_hash.get(ensobjid, None)

    def _obj_attr_subtype(self, classname: str) -> Tuple[Optional[int], Optional[dict]]:
        """Get subtype information for a given class.

        For an input class name, this method returns the proper Python proxy class name and,
        if the class supports subclasses, the attribute ID of the differentiating attribute.

        Parameters
        ----------
        classname : str
            Root class name to look up.

        Returns
        -------
        Tuple[Optional[int], Optional[dict]]
            (attr_id, subclassnamedict): Attribute used to differentiate between classes
            and a dictionary of the class names for each value of the attribute.

        """
        if classname == "ENS_PART":
            return self.ensight.objs.enums.PARTTYPE, self._subtype_tables[classname]

        elif classname == "ENS_ANNOT":
            return self.ensight.objs.enums.ANNOTTYPE, self._subtype_tables[classname]

        elif classname == "ENS_TOOL":
            return self.ensight.objs.enums.TOOLTYPE, self._subtype_tables[classname]

        return None, None

    def _convert_ctor(self, s: str) -> str:
        """Convert ENSOBJ object references into executable code in __repl__ strings.

        The __repl__() implementation for an ENSOBJ subclass generates strings like these::

            Class: ENS_GLOBALS, CvfObjID: 221, cached:yes
            Class: ENS_PART, desc: 'Sphere', CvfObjID: 1078, cached:no
            Class: ENS_PART, desc: 'engine', PartType: 0, CvfObjID: 1097, cached:no
            Class: ENS_GROUP, desc: '', Owned, CvfObjID: 1043, cached:no

        This method detects strings like those and converts them into strings like these::

            session.ensight.objs.ENS_GLOBALS(session, 221)
            session.ensight.objs.ENS_PART_MODEL(session, 1078, attr_id=1610612792, attr_value=0)

        where:

        1610612792 is ensight.objs.enums.PARTTYPE.

        If a proxy object for the ID already exists, it can also generate strings like this::

            session.obj_instance(221)


        Parameters
        ----------
        s : str
            String to convert.

        """
        self._prune_hash()
        offset = 0
        while True:
            # Find the object repl block to replace
            id = s.find("CvfObjID:", offset)
            if id == -1:
                break
            start = s.find("Class: ", offset)
            if (start == -1) or (start > id):
                break
            tail_len = 11
            tail = s.find(", cached:no", offset)
            if tail == -1:
                tail_len = 12
                tail = s.find(", cached:yes", offset)
            if tail == -1:  # pragma: no cover
                break  # pragma: no cover
            # just this object substring
            tmp = s[start + 7 : tail]
            # Subtype (PartType:, AnnotType:, ToolType:)
            subtype = None
            for name in ("PartType:", "AnnotType:", "ToolType:"):
                location = tmp.find(name)
                if location != -1:
                    subtype = int(tmp[location + len(name) :].split(",")[0])
                    break
            # Owned flag
            owned_flag = "Owned," in tmp
            # isolate the block to replace
            prefix = s[:start]
            suffix = s[tail + tail_len :]
            # parse out the object id and classname
            objid = int(s[id + 9 : tail])
            classname = s[start + 7 : tail]
            comma = classname.find(",")
            classname = classname[:comma]
            # pick the subclass based on the classname
            attr_id, classname_lookup = self._obj_attr_subtype(classname)
            # generate the replacement text
            if objid in self._ensobj_hash:
                replace_text = f"session.obj_instance({objid})"
            else:
                subclass_info = ""
                if attr_id is not None:
                    if subtype is not None:
                        # the 2024 R2 interface includes the subtype
                        if (classname_lookup is not None) and (subtype in classname_lookup):
                            classname = classname_lookup[subtype]
                            subclass_info = f",attr_id={attr_id}, attr_value={subtype}"
                    elif classname_lookup is not None:  # pragma: no cover
                        # if a "subclass" case and no subclass attrid value, ask for it...
                        remote_name = self.remote_obj(objid)
                        cmd = f"{remote_name}.getattr({attr_id})"
                        attr_value = self.cmd(cmd)
                        if attr_value in classname_lookup:
                            classname = classname_lookup[attr_value]
                            subclass_info = f",attr_id={attr_id}, attr_value={attr_value}"
                if owned_flag:
                    subclass_info += ",owned=True"
                replace_text = f"session.ensight.objs.{classname}(session, {objid}{subclass_info})"
            if replace_text is None:  # pragma: no cover
                break  # pragma: no cover
            offset = start + len(replace_text)
            s = prefix + replace_text + suffix
        s = s.strip()
        if s.startswith("[") and s.endswith("]"):
            s = f"ensobjlist({s}, session=session)"
        return s

    def capture_context(self, full_context: bool = False) -> "enscontext.EnsContext":
        """Capture the current EnSight instance state.

        This method causes the EnSight instance to save a context and return an ``EnsContext``
        object representing that saved state.

        Parameters
        ----------
        full_context : bool, optional
            Whether to include all aspects of the Ensight instance. The default is ``False``.

        Returns
        -------
        enscontext.EnsContext

        Examples
        --------
        >>> ctx = session.capture_context()
        >>> ctx.save("session_context.ctxz")

        """
        self.cmd("import ansys.pyensight.core.enscontext", do_eval=False)
        data_str = self.cmd(
            f"ansys.pyensight.core.enscontext._capture_context(ensight,{full_context})",
            do_eval=True,
        )
        context = EnsContext()
        context._from_data(data_str)
        return context

    def restore_context(self, context: "enscontext.EnsContext") -> None:
        """Restore the current EnSight instance state.

        This method restores EnSight to the state stored in an ``EnsContext``
        object that was either read from disk or returned by the
        :func:`capture_context<ansys.pyensight.core.Session.capture_context>` method.

        Parameters
        ----------
        context : enscontext.EnsContext
            Context to set the current EnSight instance to.

        Examples
        --------
        >>> tmp_ctx = session.capture_context()
        >>> session.restore_context(EnsContext("session_context.ctxz"))
        >>> session.restore_context(tmp_ctx)
        """
        data_str = context._data(b64=True)
        self.cmd("import ansys.pyensight.core.enscontext", do_eval=False)
        self.cmd(
            f"ansys.pyensight.core.enscontext._restore_context(ensight,'{data_str}')", do_eval=False
        )

    def ensight_version_check(
        self,
        version: Union[int, str],
        message: str = "",
        exception: bool = True,
        strict: bool = False,
    ) -> bool:
        """Check if the session is a specific version.

        Different versions of pyensight Sessions may host different versions of EnSight.
        This method compares the version of the remote EnSight session to a specific version
        number.  If the remote EnSight version is at least the specified version, then
        this method returns True.  If the version of EnSight is earlier than the specified
        version, this method  will raise an exception.  The caller can specify the
        error string to be included.  They may also specify if the version check should
        be for a specific version vs the specified version or higher.  It is also possible
        to avoid the exception and instead just return True or False for cases when an
        alternative implementation might be used.

        Parameters
        ----------
        version : Union[int, str]
            The version number to compare the EnSight version against.
        message : str
            The message string to be used as the text for any raised exception.
        exception : bool
            If True, and the version comparison fails, an InvalidEnSightVersion is raised.
            Otherwise, the result of the comparison is returned.
        strict : bool
            If True, the comparison of the two versions will only pass if they
            are identical.  If False, if the EnSight version is greater than or
            equal to the specified version the comparison will pass.

        Returns
        -------
            True if the comparison succeeds, False otherwise.

        Raises
        ------
            InvalidEnSightVersion if the comparison fails and exception is True.
        """
        ens_version = int(self.ensight.version("suffix"))
        # handle various input formats
        target = version
        if isinstance(target, str):  # pragma: no cover
            # could be 'year RX' or the suffix as a string
            if "R" in target:
                tmp = [int(x) for x in target.split("R")]
                target = (tmp[0] - 2000) * 10 + tmp[1]
            else:
                target = int(target)
        # check validity
        valid = ens_version == target
        at_least = ""
        if not strict:  # pragma: no cover
            at_least = "at least "
            valid = ens_version >= target
        if (not valid) and exception:
            ens_version = self.ensight.version("version-full")
            base_msg = f" ({at_least}'{version}' required, '{ens_version}' current)"
            if message:  # pragma: no cover
                message += base_msg  # pragma: no cover
            else:
                message = f"A newer version of EnSight is required to use this API:{base_msg}"
            raise InvalidEnSightVersion(message)
        return valid

    def find_remote_unused_ports(
        self,
        count: int,
        start: Optional[int] = None,
        end: Optional[int] = None,
        avoid: Optional[list[int]] = None,
    ) -> Optional[List[int]]:
        """
        Find "count" unused ports on the host system.  A port is considered
        unused if it does not respond to a "connect" attempt.  Walk the ports
        from 'start' to 'end' looking for unused ports and avoiding any ports
        in the 'avoid' list.  Stop once the desired number of ports have been
        found.  If an insufficient number of ports were found, return None.
        An admin user check is used to skip [1-1023].

        Parameters
        ----------
        count: int
            number of unused ports to find
        start: int
            the first port to check or None (random start)
        end: int
            the last port to check or None (full range check)
        avoid: list
            an optional list of ports not to check

        Returns
        -------
            the detected ports or None on failure
        """
        cmd = "from cei import find_unused_ports\n"
        cmd += f"ports = find_unused_ports({count}, start={start}, end={end}, avoid={avoid})"
        self.cmd(cmd, do_eval=False)
        return self.cmd("ports")

    def set_dsg_session(self, dsg_session: "DSGSession"):
        """Set a DSG Session for the current PyEnSight session.

        This is required if a DSGSession is running together with
        PyEnSight and the second might send gRPC requests while the first
        is blocked because the gRPC queue is full.

        Parameters
        ----------
        dsg_session: DSGSession
            a DSGSession object
        """
        self._dsg_session = dsg_session



"""Readers module.

This module contains utilities to do readers specific operations.
"""
import os
import re
from threading import Thread
import time
from types import ModuleType
from typing import Optional, Tuple, Union
import uuid

try:
    import ensight
except ImportError:
    from ansys.api.pyensight import ensight_api


class Readers:
    """A namespace to access the interfaces of specific readers"""

    def __init__(self, interface: Union["ensight_api.ensight", "ensight"]):
        self._ensight = interface
        self._dvs = DVS(self._ensight)

    @property
    def dvs(self) -> "DVS":
        """The ensight interface"""
        return self._dvs


class DVS:
    """A namespace to access specific DVS interfaces"""

    def __init__(self, interface: Union["ensight_api.ensight", "ensight"]):
        self._ensight: Union["ensight_api.ensight", "ensight"] = interface
        self._dvs_port: Optional[int] = None

    MONITOR_NEW_TIMESTEPS_STAY_AT_CURRENT = "stay_at_current"
    MONITOR_NEW_TIMESTEPS_JUMP_TO_END = "jump_to_end"

    def _launch_live_dvs_242_cmd(
        self,
        tmp_name: str,
        port: int = 0,
        secret_key: Optional[str] = None,
        monitor_new_timesteps: str = MONITOR_NEW_TIMESTEPS_STAY_AT_CURRENT,
    ):
        """Launch a live DVS session from PyEnSight for EnSight 242 and onward"""
        indent = "    "
        cmd = "import re\n"
        cmd += "def dvs_callback():\n"
        cmd += f'{indent}command_string = f"set_server_port={port}"\n'
        if secret_key is not None:
            secret_string = f'{indent}command_string += f"&set_secret_key='
            secret_string += secret_key + '"\n'
            cmd += secret_string
        tmp_name = tmp_name.replace("\\", "\\\\")
        cmd += f"{indent}reply = ensight.objs.core.CURRENTCASE[0].client_command(command_string)\n"
        cmd += f"{indent}dvs_port_string = ensight.objs.core.CURRENTCASE[0].client_command('get_server_port')\n"
        cmd += (
            f"{indent}dvs_port = int(re.search(r':([0-9]{{4,5}})', dvs_port_string).groups(1)[0])\n"
        )
        cmd += f"{indent}with open('{tmp_name}', 'w') as temp_file:\n"
        cmd += f"{2*indent}temp_file.write(str(dvs_port))\n"
        cmd += f"{indent}return True\n\n"
        cmd += "reply = ensight.objs.core.CURRENTCASE[0].client_command_callback(dvs_callback)\n"
        if monitor_new_timesteps:
            cmd += f'ensight.solution_time.monitor_for_new_steps("{monitor_new_timesteps}")\n'
        cmd += 'ensight.part.elt_representation("3D_feature_2D_full")\n'
        cmd += 'ensight.data.format("DVS")\n'
        cmd += 'err = ensight.data.replace("notexisting.dvs")\n'
        return cmd

    def _launch_live_dvs_241_cmd(
        self,
        port: int = 0,
        secret_key: Optional[str] = None,
        monitor_new_timesteps: str = MONITOR_NEW_TIMESTEPS_STAY_AT_CURRENT,
    ):
        """Launch a live DVS session from PyEnSight for EnSight 241"""
        if port == 0:
            ports = self._ensight._session.find_remote_unused_ports(1)
            if ports:
                self._dvs_port = ports[0]
                port = ports[0]
        indent = "    "
        cmd = "import os\n"
        cmd += f"try:\n{indent}os.remove('remote.dvs')\n"
        cmd += f"except (FileNotFoundError, OSError):\n{indent}pass\n"
        cmd += "path = os.path.join(os.getcwd(), 'remote.dvs')\n"
        cmd += "with open(path, 'w') as dvsfile:\n"
        cmd += f"{indent}dvsfile.write('#!DVS_CASE 1.0\\n')\n"
        cmd += f"{indent}dvsfile.write(f'SERVER_PORT_BASE={port}\\n')\n"
        cmd += f"{indent}dvsfile.write('SERVER_PORT_MULT=1\\n')\n"
        if secret_key:
            cmd += f"{indent}dvsfile.write(f'SERVER_SECURITY_SECRET={secret_key}\\n')\n"
        cmd += f'ensight.solution_time.monitor_for_new_steps("{monitor_new_timesteps}")\n'
        cmd += 'ensight.part.elt_representation("3D_feature_2D_full")\n'
        cmd += 'ensight.data.format("DVS")\n'
        cmd += "ensight.data.replace(path)\n"
        return cmd

    def _find_dvs_port(self, tmp_name: Optional[str] = None):
        """Find the dvs port allocated from the input temporary name"""
        if not tmp_name:
            raise RuntimeError("Temporary name for dvs port file not available")
        try_local = True
        if self._ensight._session._launcher:
            if hasattr(self._ensight._session._launcher, "_enshell"):
                try_local = False
                log_content = self._ensight._session._launcher.enshell_log_contents()
                dvs_port_match = re.search(r"\(0.0.0.0\):([0-9]{4,5})\n", log_content)
                if dvs_port_match:
                    self._dvs_port = int(str(dvs_port_match.groups(1)[0]).strip())
        if try_local:
            try:
                with open(tmp_name) as dvs_port_file:
                    self._dvs_port = int(dvs_port_file.read().strip())
            except Exception:
                raise RuntimeError("Cannot retrieve DVS Port")

    @staticmethod
    def _launch_dvs_callback_in_ensight(
        port: int,
        filename: Optional[str],
        secret_key: Optional[str],
        monitor_new_timesteps: str = MONITOR_NEW_TIMESTEPS_STAY_AT_CURRENT,
    ) -> None:  # pragma: no cover
        """Launch a live DVS session in EnSight"""
        from ceiversion import ensight_suffix  # pylint: disable=import-outside-toplevel

        if int(ensight_suffix) < 242:
            from cei import find_unused_ports  # pylint: disable=import-outside-toplevel

            if port == 0:
                ports = find_unused_ports(1)
                if ports:
                    port = ports[0]
            try:
                os.remove("remote.dvs")
            except (FileNotFoundError, OSError):
                pass
            path = os.path.join(os.getcwd(), "remote.dvs")
            with open(path, "w") as dvsfile:
                dvsfile.write("#!DVS_CASE 1.0\n")
                dvsfile.write(f"SERVER_PORT_BASE={port}\n")
                dvsfile.write("SERVER_PORT_MULT=1\n")
                if secret_key is not None:
                    dvsfile.write(f"SERVER_SECURITY_SECRET={secret_key}\n")
            if filename:
                try:
                    with open(filename, "w") as dvs_port_file:
                        dvs_port_file.write(str(port))
                except Exception:
                    raise RuntimeError(f"Couldn't write allocated DVS port to {filename}")
            ensight.part.elt_representation("3D_feature_2D_full")
            ensight.solution_time.monitor_for_new_steps(f"{monitor_new_timesteps}")
            ensight.data.format("DVS")
            ensight.data.replace(path)
        else:

            def dvs_callback():
                command_string = f"set_server_port={port}"
                if secret_key is not None:
                    command_string += f"&set_secret_key={secret_key}"
                ensight.objs.core.CURRENTCASE[0].client_command(command_string)
                dvs_port_string = ensight.objs.core.CURRENTCASE[0].client_command("get_server_port")
                dvs_port_match = re.search(r":([0-9]{4,5})", dvs_port_string)
                dvs_port = None
                if dvs_port_match:
                    dvs_port = int(dvs_port_match.groups(1)[0])
                if not dvs_port:
                    raise RuntimeError("DVS couldn't allocate a port")
                if filename:
                    try:
                        with open(filename, "w") as dvs_port_file:
                            dvs_port_file.write(str(dvs_port))
                    except Exception:
                        raise RuntimeError(f"Couldn't write allocated DVS port to {filename}")
                return True

            ensight.objs.core.CURRENTCASE[0].client_command_callback(dvs_callback)
            ensight.solution_time.monitor_for_new_steps(f"{monitor_new_timesteps}")
            ensight.part.elt_representation("3D_feature_2D_full")
            ensight.data.format("DVS")
            ensight.data.replace("notexisting.dvs")

    def launch_live_dvs(
        self,
        port: int = 0,
        secret_key: Optional[str] = None,
        monitor_new_timesteps: str = MONITOR_NEW_TIMESTEPS_STAY_AT_CURRENT,
        start_thread: bool = True,
        filename: Optional[str] = None,
    ) -> Tuple[Optional[Thread], Optional[int]]:
        """To provide an interface to launch an in-situ EnSight DVS session.
        If in PyEnSight, the function will return a thread which will launch the DVS reader
        in EnSight, hence the DVS servers, and will also return the port allocated by DVS, to
        cover the case the port 0 was asked for.
        If instead the function will be used directly in EnSight via the utils interface, since
        the reader launch will block the interpreter waiting for new data, the port cannot be returned
        and cannot be printed up until the first update happens.
        So, if you need to access the port number in a script and you cannot check the EnSight console,
        please supply a filename to store the dvs port into.

        Parameters
        ----------
        port: int
            the port number where the first DVS server will be started. In case of a
            SOS EnSight session, on the following server the DVS servers will be started on the
            next port, e.g. if the first server starts at port 50055, the second will start
            at port 50056 and so on
        secret_key: str
            an optional secret key to pass in case the DVS clients have been started with a secret key
            for the underlying gRPC connections. An empty string can be provided if needed
        monitor_new_timesteps: str
            set the way EnSight will monitor for new timesteps. Defaults to MONITOR_NEW_TIMESTEPS_STAY_AT_CURRENT.
            The allowed values are MONITOR_NEW_TIMESTEPS_STAY_AT_CURRENT
            and MONITOR_NEW_TIMESTEPS_JUMP_TO_END
        start_thread: bool
            True if the thread to be returned needs to be started already. Default is True
        filename: str
            An optional path to store the port number in. It will be used only if the utils is being
            called directly in EnSight.

        Returns
        -------
        Thread:
            a python Thread which holds the dvs load
        """

        def load_dvs():
            self._ensight._session.cmd(cmd, do_eval=False)

        if monitor_new_timesteps not in [
            self.MONITOR_NEW_TIMESTEPS_JUMP_TO_END,
            self.MONITOR_NEW_TIMESTEPS_STAY_AT_CURRENT,
        ]:
            raise RuntimeError(
                f"{monitor_new_timesteps} value not allowed for an in-situ DVS session"
            )
        if not isinstance(self._ensight, ModuleType):  # pragma: no cover
            self._ensight._session.ensight_version_check("2024 R1")
        else:
            self._launch_dvs_callback_in_ensight(
                port=port,
                secret_key=secret_key,
                monitor_new_timesteps=monitor_new_timesteps,
                filename=filename,
            )
            return None, None
        cmd = ""
        path = None
        if int(self._ensight._session.cei_suffix) < 242:
            cmd = self._launch_live_dvs_241_cmd(
                port=port, secret_key=secret_key, monitor_new_timesteps=monitor_new_timesteps
            )
        else:
            tmp_name = str(uuid.uuid4())
            path = os.path.join(self._ensight._session._launcher.session_directory, tmp_name)
            cmd = self._launch_live_dvs_242_cmd(
                port=port,
                secret_key=secret_key,
                monitor_new_timesteps=monitor_new_timesteps,
                tmp_name=path,
            )
        t = Thread(target=load_dvs)
        if start_thread:
            t.start()
        start = time.time()
        while not self._dvs_port and time.time() - start < 60:
            try:
                self._find_dvs_port(path)
            except Exception:
                pass
            time.sleep(0.5)
        return t, self._dvs_port



"""The module provides a concrete implementation for the DVS bindings in PyEnSight.

It also provides a collection of utilities to starts DVS servers, clients,
launch a local PyEnSight session, or connect to an existing one, and finally
to send data from the clients to the servers.
"""
import glob
import io
import logging
import os
import pathlib
import platform
import re
import sys
import tarfile
import tempfile
import threading
import time
import traceback
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union
import warnings

from ansys.api.pyensight.dvs_api import dvs_base
from ansys.pyensight.core import DockerLauncher, LocalLauncher
from ansys.pyensight.core.common import GRPC_WARNING_MESSAGE, grpc_version_check, safe_extract
import numpy

if TYPE_CHECKING:
    from ansys.pyensight.core import Session


class DVS(dvs_base):
    """Create an instance of the DVS module.

    The module tries to look for the DVS Python bindings from the input
    Ansys installation folder (which might also be the CEI folder) or the input
    lib_folder. If not found, and if a PyEnSight session is provided,
    the DVS commands will be launched on the remote EnSight Python interpreter.

    Parameters
    ----------

    session: Session
        An optional PyEnSight session. This must be provided in case the
        DVS modules needs to be used on a remote session of EnSight.
    ansys_installation: str
        The optional full path to a local Ansys installation, or the CEI folder
        from an Ansys installation
    lib_folder: str
        The optional full path to a folder that contains the DVS libraries and Python
        bindings.
    """

    def __init__(
        self,
        session: Optional["Session"] = None,
        ansys_installation: Optional[str] = None,
        lib_folder: Optional[str] = None,
    ) -> None:
        super().__init__(session=session)
        self._ansys_installation: Optional[str] = None
        if ansys_installation:
            self._ansys_installation = LocalLauncher.get_cei_install_directory(ansys_installation)
        self._lib_folder: Optional[str] = lib_folder
        if self._session:
            if not self._ansys_installation and hasattr(self._session._launcher, "_install_path"):
                self._ansys_installation = self._session._launcher._install_path
        if not self._session and not self._ansys_installation and not self._lib_folder:
            raise RuntimeError(
                "Either a PyEnSight session or an ansys installation path, or a folder containing the DVS Python modules need to be provided."
            )
        self._connect_session = self._session
        self._servers: Dict[int, Dict[str, Union[str, int]]] = {}
        self._server_ids: List[int] = []
        self._clients: Dict[int, Dict[str, Union[str, int, bool]]] = {}
        self._client_count = 0
        self._attempt_dvs_python_bindings_import()
        self._parts: Dict[int, Any] = {}
        self._vars: Dict[int, Any] = {}
        self._update_num = 0
        self._current_update = 0
        self._elem_type_map = {
            2: self.ELEMTYPE_BAR_2,
            3: self.ELEMTYPE_TRIANGLE,
            4: self.ELEMTYPE_QUADRANGLE,
        }
        self._total_ranks = 0
        if hasattr(self._session._launcher, "_install_path"):
            self._temp_cache = tempfile.TemporaryDirectory(prefix="pyensight_dvs")
            os.mkdir(os.path.join(self._temp_cache.name, "dvs_cache"))
            self._cache_folder: str = os.path.join(self._temp_cache.name, "dvs_cache")
        else:
            self._cache_folder = "/home/ensight/dvs_cache"
        self._dataset_name: Optional[str] = None
        self._secret_key: Optional[str] = None

    @staticmethod
    def _is_windows():
        """True if the platform being used is Windows."""
        return "Win" in platform.system()

    def launch_local_pyensight_session(
        self,
        use_egl=False,
        use_sos: Optional[int] = None,
        additional_command_line_options: Optional[List] = None,
        grpc_use_tcp_sockets: Optional[bool] = False,
        grpc_allow_network_connections: Optional[bool] = False,
        grpc_disable_tls: Optional[bool] = False,
        grpc_uds_pathname: Optional[str] = None,
    ):
        """Launch a local PyEnSight session.

        If an ansys installation has been provided, it will be used to launch EnSight.
        The session will be associated to the current DVS module instance.

        Parameters

        use_egl : bool, optional
            Whether to use EGL hardware for accelerated graphics. The platform
            must be able to support this hardware. This parameter is defined on
            the parent ``Launcher`` class, where the default is ``False``.
        use_sos : int, optional
            Number of EnSight servers to use for SOS (Server of Server) mode.
            This parameter is defined on the parent ``Launcher`` class, where
            the default is ``None``, in which case SOS mode is not used.
        additional_command_line_options: list, optional
            Additional command line options to be used to launch EnSight.
        grpc_use_tcp_sockets :
            If using gRPC, and if True, then allow TCP Socket based connections
            instead of only local connections.
        grpc_allow_network_connections :
            If using gRPC and using TCP Socket based connections, listen on all networks.
        grpc_disable_tls :
            If using gRPC and using TCP Socket based connections, disable TLS.
        grpc_uds_pathname :
            If using gRPC and using Unix Domain Socket based connections, explicitly
            set the pathname to the shared UDS file instead of using the default.
        """
        launcher = LocalLauncher(
            ansys_installation=self._ansys_installation,
            use_sos=use_sos,
            use_egl=use_egl,
            additional_command_line_options=additional_command_line_options,
            grpc_allow_network_connections=grpc_allow_network_connections,
            grpc_use_tcp_sockets=grpc_use_tcp_sockets,
            grpc_disable_tls=grpc_disable_tls,
            grpc_uds_pathname=grpc_uds_pathname,
        )
        session = launcher.start()
        self._session = session

    def _check_ansys_version_grpc(self):
        """Check if the gRPC security options apply to the EnSight install."""
        buildinfo_path = os.path.join(self._ansys_installation, "BUILDINFO.txt")
        if not os.path.exists(buildinfo_path):
            raise RuntimeError("Couldn't find BUILDINFO file, cannot check installation.")
        with open(buildinfo_path, "r") as buildinfo_file:
            buildinfo = buildinfo_file.read()
        version_match = re.search("Version:\s(.*)\n", buildinfo)
        internal_version_match = re.search("Internal:\s(.*)\n", buildinfo)
        if not internal_version_match:
            raise RuntimeError("Couldn't parse EnSight internal version in BUILDINFO file.")
        internal_version = internal_version_match.group(1)
        if not version_match:
            raise RuntimeError("Couldn't parse EnSight version in BUILDINFO file.")
        ensight_full_version = version_match.group(1)
        return grpc_version_check(internal_version, ensight_full_version)

    def _attempt_dvs_python_bindings_import(self):
        """Attempt to load the actual DVS Python bindings.

        If an input lib folder has been provided, it will be tried first.
        If an ansys installation has been provided, it will be tried as second choice.
        """
        if self._lib_folder:
            try:
                sys.path.append(self._lib_folder)
                import dynamic_visualization_store

                self._dvs_module = dynamic_visualization_store
            except (ModuleNotFoundError, ImportError):
                raise RuntimeError("Cannot import DVS module from provided library folder.")
        if self._ansys_installation:
            # Check if you are inside of an ansys install
            if not self._check_ansys_version_grpc():
                warnings.warn(GRPC_WARNING_MESSAGE.replace("EnSight", "DVS"))
            apex_path = glob.glob(os.path.join(self._ansys_installation, "apex???"))
            if not apex_path:
                # try dev path
                raise RuntimeError("Cannot find a valid EnSight install")
            apex_path = apex_path[-1]
            arch = "win64" if self._is_windows() else "linux_2.6_64"
            apex_libs = os.path.join(apex_path, "machines", arch)
            python_path = glob.glob(os.path.join(apex_libs, "Python-3.*"))[-1]
            apex_py_version = re.search(
                r"Python-3.([0-9]+).([0-9]+)", os.path.basename(python_path)
            )
            apex_py_major_version = apex_py_version.group(1)
            lib_path = os.path.join(python_path, "lib", f"python3.{apex_py_major_version}")
            if self._is_windows():
                lib_path = os.path.join(python_path, "DLLs")
            sys.path.append(lib_path)
            try:
                import dynamic_visualization_store

                self._dvs_module = dynamic_visualization_store
            except (ModuleNotFoundError, ImportError):
                python_cei = os.path.join(apex_libs, "Python-CEI")
                if os.path.isdir(python_cei):
                    python_cei_lib_path = os.path.join(
                        python_cei, "lib", f"python3.{apex_py_major_version}"
                    )
                    if self._is_windows():
                        python_cei_lib_path = os.path.join(python_cei, "DLLs")
                    sys.path.append(python_cei_lib_path)
                try:
                    import dynamic_visualization_store

                    self._dvs_module = dynamic_visualization_store
                except (ModuleNotFoundError, ImportError):
                    warnings.warn(
                        "Cannot import DVS module from provided ansys installation folder."
                    )

    DVS_NULL_TRANSPORT = 0
    DVS_GRPC_TRANSPORT = 1

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, session: "Session"):
        self._session = session

    def start_dvs_servers(
        self, num_servers: int, transport: int = 0, ranks_per_server: int = 1, secret_key: str = ""
    ):
        """Start DVS servers using the Python bindings.

        The DVS servers will be started externall to the eventual EnSigth session available.
        For simplicity, it is assumed that each server will receive the same number of ranks,
        declared in input.

        Parameters
        ----------
        num_servers: int
            the number of DVS servers to launch
        transport: int
            the kind of transport to be used. Defaults to null.
            Description of options as follows.

            ================== =========================================================
            Name               Query type
            ================== =========================================================
            DVS_NULL_TRANSPORT Start the servers with the null protocol. Default
            DVS_GRPC_TRANSPORT Start the servers with the grpc protocol.
            ================== =========================================================
        ranks_per_server: int
            the number or ranks that will be connected to each server. Defaults to 1
        secret_key: str
            the secret key that will be used for the eventual gRPC connection.
            Can be an empty string, that is also the default value.
        """
        if not self._secret_key:
            self._secret_key = secret_key
        transport_string = "null" if transport == 0 else "grpc"
        uri = f"{transport_string}://"
        grpc = transport == self.DVS_GRPC_TRANSPORT
        options = {"CACHE_URI": f"hdf5:///{pathlib.Path(self._cache_folder).as_posix()}"}
        if grpc:
            uri += "127.0.0.1:0"
            options.update(
                {
                    "SERVER_SECURITY_SECRET": secret_key,
                }
            )
            if self._session:
                if self._session._grpc_allow_network_connections:
                    options["SERVER_LISTEN_ALL_NETWORKS"] = "1"
                if self._session._grpc_disable_tls:
                    options["SERVER_DISABLE_TLS"] = "1"
                if self._session._grpc_use_tcp_sockets:
                    options["SERVER_USE_TCP_SOCKETS"] = "1"
                if self._session._grpc_uds_pathname:
                    options["SERVER_UNIX_DOMAIN_SOCKET_PATH"] = self._session._grpc_uds_pathname
        try:
            for n in range(0, num_servers):
                # Assume ranks equally distributed
                server_id = self.server_create(uri=uri)
                self.server_start(
                    server_id, server_num=n, local_ranks=ranks_per_server, options=options
                )
                self._server_ids.append(server_id)
                self._servers[n] = {
                    "server_id": server_id,
                    "ranks": ranks_per_server,
                    "in_ensight": False,
                }
                if grpc:
                    uri_to_save = self.server_get_uri(server_id)
                    port = re.search(":([0-9]+)", uri_to_save)
                    if port:
                        self._servers[n].update(
                            {"server_uri": uri_to_save, "port": int(port.group(1))}
                        )
            self._total_ranks = ranks_per_server * len(self._server_ids)
            started = False
            start = time.time()
            while not started and time.time() - start < 60:
                if not all([self.server_started(s) for s in self._server_ids]):
                    time.sleep(0.5)
                else:
                    started = True
            if not started:
                raise RuntimeError("The DVS servers have not started in 60 seconds.")
        except Exception as e:
            traceback.print_exc()
            raise RuntimeError(f"Couldn't start the servers, error: {e}")

    def _start_dvs_client(self, server_id: int, rank: int, dedup=False):
        """Start a DVS client.

        Parameters
        ----------
        server_id: int
            the server ID to connect to
        rank: int
            the rank of the client to launch
        secret_key: str
            the secret key to be used to connect to the DVS server
        dedup: bool
            True to not send duplicate data to server
        """
        if server_id not in self._server_ids:
            raise RuntimeError(f"Server ID {server_id} not started in this process.")
        flags = self.FLAGS_BLOCK_FOR_SERVER
        if dedup:
            flags |= self.FLAGS_DEDUP
        try:
            client_id = self.connect(server_id=server_id, secret=self._secret_key, flags=flags)
        except Exception as e:
            traceback.print_exc()
            raise RuntimeError(f"Couldn't start the client, error {e}")
        self._clients[self._client_count] = {
            "client_id": client_id,
            "server_id": server_id,
            "rank": rank,
            "update_started": False,
        }
        self._client_count += 1

    def start_dvs_servers_in_ensight(self, ranks_per_server: int, secret_key=""):
        """Launch the DVS servers in EnSight for an in-situ session.

        On each EnSight server a DVS server will be launched.

        Parameters
        ----------
        ranks_per_server: int
            how many ranks will be sent to each server. This will be used
            in a later stage for the clients launch and connection.
        secret_key: str
            the secret key that will be used for the eventual gRPC connection.
            Can be an empty string, that is also the default value.
        """
        if not self._secret_key:
            self._secret_key = secret_key
        if not self._session:
            raise RuntimeError(
                "A PyEnSight session is required to start the DVS servers in EnSight."
            )
        thread, port = self._session.ensight.utils.readers.dvs.launch_live_dvs(
            secret_key=secret_key
        )
        num_servers = self._session._launcher._use_sos or 1
        base_uri = f"grpc://{self._session.hostname}"
        self._total_ranks = num_servers * ranks_per_server
        # Need to implement SOS support in session.ensight.utils.readers.dvs.launch_live_dvs
        for n in range(num_servers):
            # Just create a server but not start it
            server_id = self.server_create(uri=base_uri + f":{port+n}")
            self._server_ids.append(server_id)
            self._servers[n] = {
                "server_uri": base_uri + ":{}".format(port + n),
                "port": port + n,
                "server_id": server_id,
                "in_ensight": True,
                "ranks": ranks_per_server,
            }

    def start_dvs_clients(self, dataset_name: str, dedup=False):
        """Launch the DVS clients and connect to the existing DVS servers.

        Parameters
        ----------
        dataset_name: str
            The dataset name required to initialize the following exports.
        """
        self._dataset_name = dataset_name
        rank_per_server = list(self._servers.values())[0].get("ranks")
        local_ranks = 0
        n = 0
        for rank in range(0, self._total_ranks):
            server = self._servers[n]
            local_ranks += 1
            if local_ranks == rank_per_server:
                local_ranks = 0
                n += 1
            self._start_dvs_client(int(server["server_id"]), rank, dedup=dedup)

    def _begin_update(
        self, client_dict: Dict[str, Union[str, int, bool]], time: float, rank: int, chunk: int
    ):
        """Start an update.

        Parameters
        ----------
        client_dict: dict
            A dictionary holding the DVS client parameters
        time: float
            The time value for the current update. May be a time already used
        rank: int
            The rank of the update
        chunk: int
            The chunk of the update
        """
        try:
            _ = self.begin_update(client_dict["client_id"], self._update_num, time, rank, chunk)
            client_dict["update_started"] = True
        except Exception as e:
            traceback.print_exc()
            raise RuntimeError(f"Couldn't begin update. Error: {e}")

    def begin_updates(self, time: float):
        """Begin an update on all the clients available for the input time value.

        Each update will be launched on a separate thread. The client associated
        to the update will be flagged for the update start.

        Currently we are assuming one chunk. Chunking support will be added in a future
        release.

        Parameters
        ----------
        time: float
            The time value for the current update. May be a time already used
        """
        for _, client_vals in self._clients.items():
            thread = threading.Thread(
                target=self._begin_update, args=(client_vals, time, client_vals["rank"], 0)
            )
            thread.start()

    def begin_initialization(self):
        """Begin initialization for all the clients."""
        for c in range(self._client_count):
            client = self._clients[c]
            _ = self.begin_init(
                client["client_id"],
                dataset_name=f"Simba_{self._dataset_name}",
                rank=client["rank"],
                total_ranks=self._total_ranks,
                num_chunks=1,
            )

    def end_initialization(self):
        """End initialization for all the clients."""
        for c in range(self._client_count):
            client = self._clients[c]
            _ = self.end_init(client["client_id"])

    def create_part(self, part_id: int, part_name: str, metadata: Optional[Dict[str, str]] = None):
        """Create a part definition for the DVS export.

        Parameters
        ----------
        part_id: int
            the ID of the part to be exported
        part_name: str
            the name of the part to export
        metadata: dict
            An optional dictionary of metadata to attach to the part.
        """
        if not metadata:
            metadata = {}
        if self._parts.get(part_id):
            print("Part already created, skip")
            return
        part = {
            "id": part_id,
            "name": part_name,
            "structured": False,
            "chunking": False,
            "tags": metadata,
        }
        for c in range(self._client_count):
            client = self._clients[c]
            self.add_part_info(client["client_id"], [part])
        self._parts[part_id] = part

    def create_variable(
        self,
        var_id: int,
        var_name: str,
        var_type: int,
        location: int,
        unit: str = "",
        unit_label="",
        metadata: Optional[Dict[str, str]] = None,
    ):
        """Create a variable definition for the DVS export.

        Parameters
        ----------
        var_id: int
            the ID of the var to be exported
        var_name: str
            the name of the var to export
        var_type: int
            The variable type. Check the VARTYPE enums available with this module
        location: int
            The variable location. Check the LOCATION enums available with this module
        unit: str
            The variable units. See https://nexusdemo.ensight.com/docs/python/html/ENS_UNITSSchema.html
        unit_label: str
            The label for the variable units. See https://nexusdemo.ensight.com/docs/python/html/ENS_UNITSSchema.html
        metadata: dict
            An optional dictionary of metadata to attach to the var.
        """
        if not metadata:
            metadata = {}
        if self._vars.get(var_id):
            print("Var already created, skip")
            return
        var = {
            "id": var_id,
            "name": var_name,
            "tags": metadata,
            "type": var_type,
            "location": location,
            "unit": unit,
            "unit_label": unit_label,
        }
        for c in range(self._client_count):
            client = self._clients[c]
            self.add_var_info(client["client_id"], [var])
        self._vars[var_id] = var

    def _check_updates_started(self):
        """Check that all the updates started successfully.

        This is required because the launch of the updates is threaded.
        """
        started = False
        start = time.time()
        while not started and time.time() - start < 60:
            started = all([vals["update_started"] for c, vals in self._clients.items()])
            if not started:
                time.sleep(0.5)
        if not started:
            for c, vals in self._clients.items():
                update = vals["update_started"]
                logging.debug(f"Client {c}, update: {update}")
            raise RuntimeError("Not all clients have begun the updates.")

    def send_coordinates(self, part_id: int, vertices: Union[List[float], numpy.ndarray]):
        """Send the coordinates data for the input part.

        The full coordinates array will be sent across all the ranks.
        The data will be used for building a mesh chunk in DVS.
        The data are assumed in the following format:
        [x0, y0, z0, x1, y1, z1, ...]

        Parameters
        ----------
        part_id: int
            the part to define the coordinates for
        vertices: List[int] or numpy array
            the coordinates array. The format is described above.
        """
        if not self._parts.get(part_id):
            raise RuntimeError(
                "Please create the part first via create_part() or the lower level add_part_info."
            )
        if not isinstance(vertices, numpy.ndarray):
            vertices = numpy.array(vertices)
        reshaped_vertices = vertices.reshape(-1, 3)
        x_coords = reshaped_vertices[:, 0]
        y_coords = reshaped_vertices[:, 1]
        z_coords = reshaped_vertices[:, 2]
        self._check_updates_started()
        for c in range(self._client_count):
            client = self._clients[c]
            self.update_nodes(
                client["client_id"], part_id=part_id, x=x_coords, y=y_coords, z=z_coords
            )

    def send_variable_data(
        self,
        var_id: int,
        part_id: int,
        values: Union[List[float], numpy.ndarray],
    ):
        """Send the variable data for the input variable.

        Parameters
        ----------
        var_id: int
            the ID of the variable that will get its values updated.
        part_id: int
            the ID of the part to update the variable for
        values: List[int] or numpy array
            the variablle array. If the variable is a vector, the values are expected as
            [v1x, v1y, v1z, v2x, v2y, v2z ...]
        """
        if not self._vars.get(var_id):
            raise RuntimeError(
                "Please create the variable first via create_var() or the lower level add_var_info."
            )
        if not self._parts.get(part_id):
            raise RuntimeError(
                "Please create the part first via create_part() or the lower level add_part_info."
            )
        if not self._parts[part_id].get("dvs_elem_type"):
            raise RuntimeError(f"Please send first the part connectivity for part {part_id}")
        elem_type = self._parts[part_id]["dvs_elem_type"]
        if not isinstance(values, numpy.ndarray):
            values = numpy.array(values)
        self._check_updates_started()
        var_type = self._vars[var_id]["type"]
        location = self._vars[var_id]["location"]
        # The following checks are there just to make mypy happy
        if isinstance(var_type, (str, bool, dict)):
            raise RuntimeError("Var type is not an integer")
        if isinstance(location, (str, bool, dict)):
            raise RuntimeError("Location is not an integer")
        for c in range(self._client_count):
            client = self._clients[c]
            if var_type == self.VARTYPE_SCALAR:
                if location == self.LOCATION_NODE:
                    self.update_var_node_scalar(
                        client["client_id"], var_id=var_id, part_id=part_id, values=values
                    )
                elif location == self.LOCATION_ELEMENT:
                    self.update_var_element_scalar(
                        client["client_id"],
                        var_id=var_id,
                        part_id=part_id,
                        elem_type=elem_type,
                        values=values,
                    )
            elif var_type == self.VARTYPE_VECTOR:
                if location == self.LOCATION_NODE:
                    self.update_var_node_vector(
                        client["client_id"], var_id=var_id, part_id=part_id, values=values
                    )
                elif location == self.LOCATION_ELEMENT:
                    self.update_var_element_vector(
                        client["client_id"],
                        var_id=var_id,
                        part_id=part_id,
                        elem_type=elem_type,
                        values=values,
                    )

    @staticmethod
    def _split_list(lst: Union[List[int], List[float]], num_parts: int):
        """Split the input list in n parts.

        lst: list
            the list to be split
        num_parts: int
            the number of parts to split the list into

        Returns
        -------
        parts: list
            A list containing the parts the original list was split into
        """
        n = len(lst)
        part_size = n // num_parts
        remainder = n % num_parts
        parts = []
        start = 0
        for i in range(num_parts):
            end = start + part_size + (1 if i < remainder else 0)
            parts.append(lst[start:end])
            start = end
        return parts

    def send_connectivity(
        self,
        part_id,
        offsets: Union[List, numpy.ndarray],
        faces: Union[List, numpy.ndarray],
        ghost=False,
    ):
        """Send the connectivity data for the input part.

        The data will be used for building an element block in DVS.
        The connectivity array will be split among all the available ranks.
        The faces data are assumed in the following format:
        [n, i1, i2, ...in, m, j1, j2, ...jn, p, k1, k2, ...kp, ...]
        The offsets data instead:
        [0, n, n+m, n+m+p ....]
        The faces list indicates the IDs of the vertices of each face, in order.
        The offsets lists indicates the index where to find a specific face.

        Parameters
        ----------
        part_id: int
            the part to define the connectivity for
        offsets: List[int] or numpy array
            the offsets values. The format is described above.
        faces: List[int] or numpy array
            the connectivity value. The format is described above.
        ghost: bool
            True if the input data contains ghost elements.
        """
        if not self._clients:
            raise RuntimeError("No DVS clients started yet.")
        if not self._parts.get(part_id):
            raise RuntimeError(
                "Please create the part first via create_part() or the lower level add_part_info."
            )
        if not isinstance(faces, numpy.ndarray):
            faces = numpy.array(faces)
        if not isinstance(offsets, numpy.ndarray):
            offsets = numpy.array(offsets)
        vertices_per_face = numpy.diff(offsets)
        connectivity_split = numpy.split(faces, numpy.cumsum(vertices_per_face[:-1]))
        elem_type = self.ELEMTYPE_N_SIDED_POLYGON
        all_same = numpy.all(numpy.array(vertices_per_face) == vertices_per_face[0])
        if all_same:
            num_vertices = vertices_per_face[0]
            _elem_type = self._elem_type_map.get(num_vertices)
            if _elem_type:
                elem_type = _elem_type
        if ghost:
            elem_type += 1
        self._check_updates_started()
        split_arrays = self._split_list(connectivity_split, self._total_ranks)
        split_num_faces = self._split_list(vertices_per_face, self._total_ranks)
        for c in range(self._client_count):
            client = self._clients[c]
            arrays = split_arrays[c]
            if len(arrays) > 1:
                indices = numpy.concatenate(arrays)
            elif arrays:
                indices = arrays[0]
            else:
                indices = numpy.array([])
            if elem_type not in [
                self.ELEMTYPE_N_SIDED_POLYGON,
                self.ELEMTYPE_N_SIDED_POLYGON_GHOST,
            ]:
                self.update_elements(
                    client["client_id"], part_id=part_id, elem_type=elem_type, indices=indices
                )
            else:
                connectivity_num_faces = split_num_faces[c]
                self.update_elements_polygon(
                    client["client_id"],
                    part_id=part_id,
                    elem_type=elem_type,
                    nodes_per_polygon=numpy.array(connectivity_num_faces),
                    indices=indices,
                )
        self._parts[part_id]["dvs_elem_type"] = elem_type

    def _check_timestep_count(self, timeout=120.0):
        """Check that there are no pending timesteps before loading data.

        Parameters
        ----------
        timeout: float
            the timeout to set while checking for pending timesteps
        """
        ready = False
        start = time.time()
        while not ready and time.time() - start < timeout:
            vals = []
            for server_id in self._server_ids:
                num_pending, num_complete = self.server_timestep_count(server_id)
                vals.append(num_pending == 0)
            ready = all(vals)
            if not ready:
                time.sleep(0.5)
        if not ready:
            raise RuntimeError(
                f"There are still pending timesteps within the input timeout of {timeout} seconds"
            )

    def load_dataset_in_ensight(self, timeout=120.0):
        """Launch the cached dataset in EnSight.

        Parameters
        ----------
        timeout: float
            the timeout to set while checking for pending timesteps
        """
        if not self._session:
            raise RuntimeError("A PyEnSight session must be available.")
        self._check_timestep_count(timeout=timeout)
        self._session.load_data(os.path.join(self._cache_folder, f"Simba_{self._dataset_name}.dvs"))

    def end_updates(self):
        """End the current updates."""
        for c in range(self._client_count):
            client = self._clients[c]
            _ = self.end_update(client["client_id"])
            client["update_started"] = False
        self._update_num += 1

    def delete_item_on_clients(self, update_num, filter=""):
        """Delete an item from all the running clients.

        Parameters
        ----------
        update_num: int
            the update number to remove from the database
        filter: str
            the filter to apply when deleting the update number
        """
        for c in range(self._client_count):
            client = self._clients[c]
            _ = self.delete_item(client["client_id"], update_num, client["rank"], filter)

    def get_dvs_data_from_container(self, destination: str, use_docker=False):
        """Utility to save the data from the container to a local destination.

        destination: str
            the folder where to copy the files to
        use_docker: bool
            if True, download is done using the docker CLI
        """
        if not isinstance(self._session._launcher, DockerLauncher):
            raise RuntimeError("Method only available for DockerLauncher instances.")
        if not os.path.exists(destination):
            os.makedirs(destination)
        posix_uri = pathlib.Path(destination).as_uri()
        if use_docker:
            bits, stat = self._session._launcher._container.get_archive(self._cache_folder)
            with tarfile.open(fileobj=io.BytesIO(b"".join(bits)), mode="r") as tar:
                safe_extract(tar, destination)
            os.remove(bits)
        else:
            self._session.copy_from_session(posix_uri, ["dvs_cache"])



"""ensight_grpc module

This package defines the EnSightGRPC class which provides a simpler
interface to the EnSight gRPC interface, including event streams.

"""
from concurrent import futures
import os
import platform
import sys
import tempfile
import threading
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Tuple, Union
import uuid

from ansys.api.pyensight.v0 import dynamic_scene_graph_pb2_grpc, ensight_pb2, ensight_pb2_grpc
from ansys.tools.common.cyberchannel import create_channel
import grpc

if TYPE_CHECKING:
    from ansys.pyensight.core import Session
    from ansys.pyensight.core.utils.dsg_server import DSGSession


class EnSightGRPC(object):
    """Wrapper around a gRPC connection to an EnSight instance

    This class provides an asynchronous interface to the EnSight
    core gRPC interface.  It can handle remote event
    streams, providing a much simpler interface to the EnSight
    application. The default is to make a connection to an EnSight
    gRPC server on port 12345 on the loopback host.

    Parameters
    ----------
    host: str, optional
        Hostname where there EnSight gRPC server is running.
    port: int, optional
        Port to make the gRPC connection to
    secret_key: str, optional
        Connection secret key
    grpc_use_tcp_sockets: bool, optional
        If using gRPC, and if True, then allow TCP Socket based connections
        instead of only local connections.
    grpc_allow_network_connections: bool, optional
        If using gRPC and using TCP Socket based connections, listen on all networks.
    grpc_disable_tls: bool, optional
        If using gRPC and using TCP Socket based connections, disable TLS.
    grpc_uds_pathname: str, optional
        If using gRPC and using Unix Domain Socket based connections, explicitly
        set the pathname to the shared UDS file instead of using the default.
    session: Session
        The PyEnSight session for this connection. This is optional, and if set
        it is used to check if the current install can handle the gRPC security options.
    WARNING:
    Overriding the default values for these options: grpc_use_tcp_sockets, grpc_allow_network_connections,
    and grpc_disable_tls
    can possibly permit control of this computer and any data which resides on it.
    Modification of this configuration is not recommended.  Please see the
    documentation for your installed product for additional information.
    """

    def __init__(
        self,
        host: str = "127.0.0.1",
        port: int = 12345,
        secret_key: str = "",
        grpc_use_tcp_sockets: bool = False,
        grpc_allow_network_connections: bool = False,
        grpc_disable_tls: bool = False,
        grpc_uds_pathname: Optional[str] = None,
        session: Optional["Session"] = None,
    ):
        self._host = host
        self._port = port
        self._channel = None
        self._stub = None
        self._dsg_stub = None
        self._security_token = secret_key
        self._grpc_use_tcp_sockets = grpc_use_tcp_sockets
        self._grpc_allow_network_connections = grpc_allow_network_connections
        self._grpc_disable_tls = grpc_disable_tls
        self._grpc_uds_pathname = grpc_uds_pathname
        self._session_name: str = ""
        # Streaming APIs
        # Event (strings)
        self._event_stream = None
        self._event_thread: Optional[threading.Thread] = None
        self._events: List[Any] = list()
        # Callback for events (self._events not used)
        self._event_callback: Optional[Callable] = None
        self._prefix: Optional[str] = None
        self._shmem_module = None
        self._shmem_filename: Optional[str] = None
        self._shmem_client = None
        self._image_stream = None
        self._image_thread = None
        self._image = None
        self._image_number = 0
        self._sub_service = None
        self._dsg_session: Optional["DSGSession"] = None
        self._pyensight_session = session

    def set_dsg_session(self, dsg_session: "DSGSession"):
        self._dsg_session = dsg_session

    @property
    def host(self) -> str:
        """The gRPC server (EnSight) hostname"""
        return self._host

    def port(self) -> int:
        """The gRPC server (EnSight) port number"""
        return self._port

    @property
    def security_token(self) -> str:
        """The gRPC server (EnSight) secret key

        EnSight supports a security token in either numeric (-security {int}) or
        string (ENSIGHT_SECURITY_TOKEN environmental variable) form.  If EnSight
        is using a security token, all gRPC calls must include this token.  This
        call sets the token for all grPC calls made by this class.
        """
        return self._security_token

    @security_token.setter
    def security_token(self, name: str) -> None:
        self._security_token = name  # pragma: no cover

    @property
    def grpc_use_tcp_sockets(self) -> bool:
        """Get whether to use Unix Domain Sockets or TCP Sockets for gRPC"""
        return self._grpc_use_tcp_sockets

    @grpc_use_tcp_sockets.setter
    def grpc_use_tcp_sockets(self, use_sockets: bool) -> None:
        """Set whether to use Unix Domain Sockets or TCP Sockets for gRPC"""
        self._grpc_use_tcp_sockets = use_sockets

    @property
    def grpc_allow_network_connections(self) -> bool:
        """Get whether to allow listening on all networks if using TCP Sockets for gRPC"""
        return self._grpc_use_tcp_sockets

    @grpc_allow_network_connections.setter
    def grpc_allow_network_connections(self, allow: bool) -> None:
        """Set whether to allow listening on all networks if using TCP Sockets for gRPC"""
        self._grpc_allow_network_connections = allow

    @property
    def grpc_disable_tls(self) -> bool:
        """Get whether to use TLS for TCP Sockets for gRPC"""
        return self._grpc_disable_tls

    @grpc_disable_tls.setter
    def grpc_disable_tls(self, disable_tls: bool) -> None:
        """Set whether to use TLS for TCP Sockets for gRPC"""
        self._grpc_disable_tls = disable_tls

    @property
    def grpc_uds_pathname(self) -> Optional[str]:
        """Get the pathname for the UDS file if not using the default for gRPC"""
        return self._grpc_uds_pathname

    @grpc_uds_pathname.setter
    def grpc_uds_pathname(self, uds_pathname: str) -> None:
        """Set the pathname for the UDS file if not using the default for gRPC"""
        self._grpc_uds_pathname = uds_pathname

    @property
    def session_name(self) -> str:
        """The gRPC server session name

        EnSight gRPC calls can include the session name via 'session_name' metadata.
        A client session may provide a session name via this property.
        """
        return self._session_name

    @session_name.setter
    def session_name(self, name: str) -> None:
        self._session_name = name

    def shutdown(self, stop_ensight: bool = False, force: bool = False) -> None:
        """Close down the gRPC connection

        Disconnect all connections to the gRPC server.  If stop_ensight is True, send the
        'Exit' command to the EnSight gRPC server.

        Parameters
        ----------
        stop_ensight: bool, optional
            if True, send an 'Exit' command to the gRPC server.
        force: bool, optional
            if stop_ensight and force are true, stop EnSight aggressively
        """
        if self.is_connected():  # pragma: no cover
            # if requested, send 'Exit'
            if stop_ensight:  # pragma: no cover
                # the gRPC ExitRequest is exactly that, a request in some
                # cases the operation needs to be forced
                if force:  # pragma: no cover
                    try:
                        self.command("ensight.exit(0)", do_eval=False)
                    except IOError:  # pragma: no cover
                        # we expect this as the exit can result in the gRPC call failing
                        pass  # pragma: no cover
                else:
                    if self._stub:  # pragma: no cover
                        _ = self._stub.Exit(
                            ensight_pb2.ExitRequest(), metadata=self._metadata()
                        )  # pragma: no cover
            # clean up control objects
            self._stub = None
            self._dsg_stub = None
            if self._channel:
                self._channel.close()
            self._channel = None
            if self._shmem_client:
                if self._shmem_module:
                    self._shmem_module.stream_destroy(self._shmem_client)
                else:
                    self.command("ensight_grpc_shmem.stream_destroy(enscl._shmem_client)")
                self._shmem_client = None

    def is_connected(self) -> bool:
        """Check to see if the gRPC connection is live

        Returns
        -------
             True if the connection is active.
        """
        return self._channel is not None

    def connect(self, timeout: float = 15.0) -> None:
        """Establish the gRPC connection to EnSight

        Attempt to connect to an EnSight gRPC server using the host and port
        established by the constructor.  Note on failure, this function just
        returns, but is_connected() will return False.

        Parameters
        ----------
        timeout: float
            how long to wait for the connection to timeout
        """
        if self.is_connected():
            return
        # set up the channel
        transport_mode = None
        host = None
        port = None
        uds_service = None
        uds_dir = None
        options = [
            ("grpc.max_receive_message_length", -1),
            ("grpc.max_send_message_length", -1),
            ("grpc.testing.fixed_reconnect_backoff_ms", 1100),
        ]
        if self._grpc_use_tcp_sockets:
            host = self._host
            transport_mode = "mtls"
            if self._grpc_disable_tls:
                transport_mode = "insecure"
            port = self._port
        else:
            host = "127.0.0.1"
            if sys.platform == "win32":
                transport_mode = "wnua"
                port = self._port
            else:
                transport_mode = "uds"
                if self._grpc_uds_pathname:
                    uds_service = os.path.basename(self._grpc_uds_pathname)
                    uds_dir = os.path.dirname(self._grpc_uds_pathname)
                else:
                    uds_dir = "/tmp"
                    uds_service = "greeter"
        # Ignore the security options if the version of EnSight cannot handle them
        if self._pyensight_session and self._pyensight_session._launcher:
            if not self._pyensight_session._launcher._has_grpc_changes:
                transport_mode = "insecure"
                host = self._host
                port = self._port
        self._channel = create_channel(
            host=host,
            port=port,
            transport_mode=transport_mode,
            uds_dir=uds_dir,
            uds_service=uds_service,
            grpc_options=options,
        )
        try:
            grpc.channel_ready_future(self._channel).result(timeout=timeout)
        except grpc.FutureTimeoutError:  # pragma: no cover
            self._channel = None  # pragma: no cover
            return  # pragma: no cover
        # hook up the stub interface
        self._stub = ensight_pb2_grpc.EnSightServiceStub(self._channel)
        self._dsg_stub = dynamic_scene_graph_pb2_grpc.DynamicSceneGraphServiceStub(self._channel)

    def _metadata(self) -> List[Tuple[bytes, Union[str, bytes]]]:
        """Compute the gRPC stream metadata

        Compute the list to be passed to the gRPC calls for things like security
        and the session name.

        """
        ret: List[Tuple[bytes, Union[str, bytes]]] = list()
        s: Union[str, bytes]
        if self._security_token:  # pragma: no cover
            s = self._security_token
            if type(s) == str:  # pragma: no cover
                s = s.encode("utf-8")
            ret.append((b"shared_secret", s))
        if self.session_name:  # pragma: no cover
            s = self.session_name.encode("utf-8")
            ret.append((b"session_name", s))
        return ret

    def render(
        self,
        width: int = 640,
        height: int = 480,
        aa: int = 1,
        png: bool = True,
        highlighting: bool = False,
    ) -> bytes:
        """Generate a rendering of the current EnSight scene

        Render the current scene at a specific size and using a specific number of anti-aliasing
        passes. The return value can be a byte array (width*height*3) bytes or a PNG image.

        Parameters
        ----------
        width: int, optional
            width of the image to render
        height: int, optional
            height of the image to render
        aa: int, optional
            number of antialiasing passes to use in generating the image
        png: bool, optional
            if True, the return value is a PNG image bytestream.  Otherwise, it is a simple
            bytes object with width*height*3 values.
        highlighting: bool, optional
            if True, selection highlighting will be included in the image.

        Returns
        -------
        bytes
            bytes object representation of the rendered image

        Raises
        ------
            IOError if the operation fails
        """
        self.connect()
        ret_type = ensight_pb2.RenderRequest.IMAGE_RAW
        if png:  # pragma: no cover
            ret_type = ensight_pb2.RenderRequest.IMAGE_PNG
        response: Any
        try:
            if self._stub:  # pragma: no cover
                response = self._stub.RenderImage(
                    ensight_pb2.RenderRequest(
                        type=ret_type,
                        image_width=width,
                        image_height=height,
                        image_aa_passes=aa,
                        include_highlighting=highlighting,
                    ),
                    metadata=self._metadata(),
                )
        except Exception:  # pragma: no cover
            raise IOError("gRPC connection dropped")  # pragma: no cover
        return response.value

    def geometry(self) -> bytes:
        """Return the current scene geometry in glTF format

        Package up the geometry currently being viewed in the EnSight session as
        a glTF stream.  Return this stream as an array of byte.  Note: no
        intermediate files are utilized.

        Note: currently there is a limitation of glTF files to 2GB

        Returns
        -------
            bytes object representation of the glTF file

        Raises
        ------
            IOError if the operation fails
        """
        self.connect()
        response: Any
        try:
            if self._stub:  # pragma: no cover
                response = self._stub.GetGeometry(
                    ensight_pb2.GeometryRequest(type=ensight_pb2.GeometryRequest.GEOMETRY_GLB),
                    metadata=self._metadata(),
                )
        except Exception:  # pragma: no cover
            raise IOError("gRPC connection dropped")  # pragma: no cover
        return response.value

    def command(self, command_string: str, do_eval: bool = True, json: bool = False) -> Any:
        """Send a Python command string to be executed in EnSight

        The string will be run or evaluated in the EnSight Python interpreter via the
        EnSightService::RunPython() gRPC all.  If an exception or other error occurs, this
        function will throw a RuntimeError.  If do_eval is False, the return value will be None,
        otherwise it will be the returned string (eval() will not be performed).  If json is True,
        the return value will be a JSON representation of the report execution result.

        Parameters
        ----------
        command_string: str
            The string to execute
        do_eval: bool, optional
            If True, a return value will be computed and returned
        json: bool, optional
            If True and do_eval is True, the return value will be a JSON representation of
            the evaluated value.

        Returns
        -------
        Any
             None, a string ready for Python eval() or a JSON string.

        Raises
        ------
            RuntimeError if the operation fails.
            IOError if the communication fails.
        """
        self.connect()
        flags = ensight_pb2.PythonRequest.EXEC_RETURN_PYTHON
        response: Any
        if json:  # pragma: no cover
            flags = ensight_pb2.PythonRequest.EXEC_RETURN_JSON  # pragma: no cover
        if not do_eval:
            flags = ensight_pb2.PythonRequest.EXEC_NO_RESULT
        try:
            if self._stub:  # pragma: no cover
                response = self._stub.RunPython(
                    ensight_pb2.PythonRequest(type=flags, command=command_string),
                    metadata=self._metadata(),
                )
        except Exception:
            raise IOError("gRPC connection dropped")
        if response.error < 0:  # pragma: no cover
            raise RuntimeError(response.value)  # pragma: no cover
        if flags == ensight_pb2.PythonRequest.EXEC_NO_RESULT:
            return None
        # This was moved externally so pre-processing could be performed
        # elif flags == ensight_pb2.PythonRequest.EXEC_RETURN_PYTHON:
        #    return eval(response.value)
        return response.value

    def prefix(self) -> str:
        """Return the unique prefix for this instance.

        Some EnSight gRPC APIs require a unique prefix so that EnSight can handle
        multiple, simultaneous remote connections. This method will generate a GUID-based
        prefix.

        Returns
        -------
        str
            A unique (for this session) prefix string of the form: grpc://{uuid}/
        """
        # prefix URIs will have the format:  "grpc://{uuid}/{callbackname}?enum={}&uid={}"
        if self._prefix is None:
            self._prefix = "grpc://" + str(uuid.uuid1()) + "/"
        return self._prefix

    def event_stream_enable(self, callback: Optional[Callable] = None) -> None:
        """Enable a simple gRPC-based event stream from EnSight

        This method makes a EnSightService::GetEventStream() gRPC call into EnSight, returning
        an ensightservice::EventReply stream.  The method creates a thread to hold this
        stream open and read new events from it.  The thread adds the event strings to
        a list of events stored on this instance.  If callback is not None, the object
        will be called with the event string, otherwise they can be retrieved using get_event().
        """
        if self._event_stream is not None:  # pragma: no cover
            return  # pragma: no cover
        self._event_callback = callback
        self.connect()
        if self._stub:  # pragma: no cover
            self._event_stream = self._stub.GetEventStream(
                ensight_pb2.EventStreamRequest(prefix=self.prefix()),
                metadata=self._metadata(),
            )
        self._event_thread = threading.Thread(target=self._poll_events)
        self._event_thread.daemon = True
        self._event_thread.start()

    def event_stream_is_enabled(self) -> bool:
        """Check to see if the event stream is enabled

        If an event stream has been successfully established via
        event_stream_enable(), then this function returns True.

        Returns
        -------
              True if a ensightservice::EventReply steam is active
        """
        return self._event_stream is not None  # pragma: no cover

    def dynamic_scene_graph_stream(self, client_cmds):  # pragma: no cover
        """Open up a dynamic scene graph stream

        Make a DynamicSceneGraphService::GetSceneStream() rpc call and return
        a ensightservice::SceneUpdateCommand stream instance.

        Parameters
        ----------
        client_cmds
            iterator that produces ensightservice::SceneClientCommand objects

        Returns
        -------
            ensightservice::SceneUpdateCommand stream instance
        """
        self.connect()
        return self._dsg_stub.GetSceneStream(client_cmds, metadata=self._metadata())

    def get_event(self) -> Optional[str]:  # pragma: no cover
        """Retrieve and remove the oldest ensightservice::EventReply string

        When any of the event streaming systems is enabled, Python threads will receive the
        event records and store them in this instance in an ordered fashion.  This method
        retrieves the oldest ensightservice::EventReply string in the queue.

        Returns
        -------
            None or the oldest event string in the queue.
        """
        try:
            return self._events.pop(0)
        except IndexError:
            return None

    def _put_event(self, evt: "ensight_pb2.EventReply") -> None:
        """Add an event record to the event queue on this instance

        This method is used by threads to make the events they receive available to
        calling applications via get_event().
        """
        if self._event_callback:  # pragma: no cover
            self._event_callback(evt.tag)
            return
        self._events.append(evt.tag)  # pragma: no cover

    def _poll_events(self) -> None:
        """Internal method to handle event streams

        This method is called by a Python thread to read events via the established
        ensightservice::EventReply stream.
        """
        try:
            while self._stub is not None:  # pragma: no cover
                evt = self._event_stream.next()
                self._put_event(evt)
        except Exception:
            # signal that the gRPC connection has broken
            self._event_stream = None
            self._event_thread = None

    def _attempt_shared_mem_import(self):
        try:
            import ensight_grpc_shmem

            self._shmem_module = ensight_grpc_shmem
        except ModuleNotFoundError:
            try:
                self.command("import enve", do_eval=False)
                cei_home = eval(self.command("enve.home()"))
                self.command("import ceiversion", do_eval=False)
                cei_version = eval(self.command("ceiversion.version_suffix"))
                self.command("import sys", do_eval=False)
                py_version = eval(self.command("sys.version_info[:3]"))
                is_win = True if "Win" in platform.system() else False
                plat = "win64" if is_win else "linux_2.6_64"
                _lib = "DLLs" if is_win else f"lib/python{py_version[0]}.{py_version[1]}"
                dll_loc = os.path.join(
                    cei_home,
                    f"apex{cei_version}",
                    "machines",
                    plat,
                    f"Python-{py_version[0]}.{py_version[1]}.{py_version[2]}",
                    _lib,
                )
                if os.path.exists(dll_loc):
                    sys.path.append(dll_loc)
                    import ensight_grpc_shmem

                    self._shmem_module = ensight_grpc_shmem
            except ModuleNotFoundError:
                pass

    @classmethod
    def _find_filename(cls, size=1024 * 1024 * 25):
        """Create a file on disk to support shared memory transport.

        A file, 25MB in size, will be created using the pid of the current
        process to generate the filename. It will be located in a temporary
        directory.
        """
        tempdir = tempfile.mkdtemp(prefix="pyensight_shmem")
        for i in range(100):
            filename = os.path.join(tempdir, "shmem_{}.bin".format(os.getpid() + i))
            if not os.path.exists(filename):
                try:
                    tmp = open(filename, "wb")
                    tmp.write(b"\0" * size)  # 25MB
                    tmp.close()
                    return filename
                except Exception:
                    pass
        return None

    def get_image(self):
        """Retrieve the current EnSight image.

        When any of the image streaming systems is enabled, Python threads will receive the
        most recent image and store them in this instance.  The frame stored in this instance
        can be accessed by calling this method

        Returns
        -------
        (tuple):
            A tuple containing a dictionary defining the image binary
            (pixels=bytearray, width=w, height=h) and the image frame number.
        """
        return self._image, self._image_number

    def _start_sub_service(self):
        """Start a gRPC client service.
        When the client calls one subscribe_events() or subscribe_images() with the
        connection set to GRPC, the interface requires the client to start a gRPC server
        that EnSight will call back to with event/image messages.  This method starts
        such a gRPC server."""
        try:
            if self._sub_service is not None:
                return
            self._sub_service = _EnSightSubServicer(parent=self)
            self._sub_service.start()
        except Exception:
            self._sub_service = None

    def subscribe_images(self, flip_vertical=False, use_shmem=True):
        """Subscribe to an image stream.

        This methond makes a EnSightService::SubscribeImages() gRPC call.  If
        use_shmem is False, the transport system will be made over gRPC.  It causes
        EnSight to make a reverse gRPC connection over with gRPC calls with the
        various images will be made.  If use_shmem is True (the default), the \ref shmem will be used.

        Parameters
        ---------
        flip_vertical: bool
            If True, the image pixels will be flipped over the X axis
        use_shmem: bool
            If True, use the shared memory transport, otherwise use reverse gRPC"""
        self.connect()
        if use_shmem:
            try:
                # we need a shared memory file
                self._shmem_filename = self._find_filename()
                if self._shmem_filename is not None:
                    conn_type = ensight_pb2.SubscribeImageOptions.SHARED_MEM
                    options = dict(filename=self._shmem_filename)
                    image_options = ensight_pb2.SubscribeImageOptions(
                        prefix=self.prefix(),
                        type=conn_type,
                        options=options,
                        flip_vertical=flip_vertical,
                        chunk=False,
                    )
                    _ = self._stub.SubscribeImages(image_options, metadata=self._metadata())
                    # start the local server
                    if not self._shmem_module:
                        self._attempt_shared_mem_import()
                    if self._shmem_module:
                        self._shmem_client = self._shmem_module.stream_create(self._shmem_filename)
                    else:
                        self.command("import ensight_grpc_shmem", do_eval=False)
                        to_send = self._shmem_filename.replace("\\", "\\\\")
                        self.command(
                            f"enscl._shmem_client = ensight_grpc_shmem.stream_create('{to_send}')",
                            do_eval=False,
                        )
                        if self.command("enscl._shmem_client is not None"):
                            self._shmem_client = True

                    # turn on the polling thread
                    self._image_thread = threading.Thread(target=self._poll_images)
                    self._image_thread.daemon = True
                    self._image_thread.start()
                    return
            except Exception as e:
                print("Unable to subscribe to an image stream via shared memory: {}".format(str(e)))

        self._start_sub_service()
        conn_type = ensight_pb2.SubscribeImageOptions.GRPC
        options = {}
        if self._sub_service:
            options = dict(uri=self._sub_service._uri)
        image_options = ensight_pb2.SubscribeImageOptions(
            prefix=self.prefix(),
            type=conn_type,
            options=options,
            flip_vertical=flip_vertical,
            chunk=True,
        )
        _ = self._stub.SubscribeImages(image_options, metadata=self._metadata())

    def image_stream_enable(self, flip_vertical=False):
        """Enable a simple gRPC-based image stream from EnSight.

        This method makes a EnSightService::GetImageStream() gRPC call into EnSight, returning
        an ensightservice::ImageReply stream.  The method creates a thread to hold this
        stream open and read new image frames from it.  The thread places the read images
        in this object.  An external application can retrieve the most recent one using
        get_image().

        Parameters
        ----------
        flip_vertical: bool
            If True, the image will be flipped over the X axis before being sent from EnSight."""
        if self._image_stream is not None:
            return
        self.connect()
        self._image_stream = self._stub.GetImageStream(
            ensight_pb2.ImageStreamRequest(flip_vertical=flip_vertical, chunk=True),
            metadata=self._metadata(),
        )
        self._image_thread = threading.Thread(target=self._poll_images)
        self._image_thread.daemon = True
        self._image_thread.start()

    def _put_image(self, the_image):
        """Store an image on this instance.

        This method is used by threads to store the latest image they receive
        so it can be accessed by get_image.
        """
        self._image = the_image
        self._image_number += 1

    def image_stream_is_enabled(self):
        """Check to see if the image stream is enabled.

        If an image stream has been successfully established via image_stream_enable(),
        then this function returns True.

        Returns
        -------
        (bool):
            True if a ensightservice::ImageReply steam is active
        """
        return self._image_stream is not None

    def _poll_images(self):
        """Handle image streams.

        This method is called by a Python thread to read imagery via the shared memory
        transport system or the the ensightservice::ImageReply stream.
        """
        try:
            while self._stub is not None:
                if self._shmem_client:
                    if self._shmem_module:
                        img = self._shmem_module.stream_lock(self._shmem_client)
                    else:
                        img = self.command("ensight_grpc_shmem.stream_lock(enscl._shmem_client)")
                    if type(img) is dict:
                        the_image = dict(
                            pixels=img["pixeldata"], width=img["width"], height=img["height"]
                        )
                        self._put_image(the_image)
                        if self._shmem_module:
                            self._shmem_module.stream_unlock(self._shmem_client)
                        else:
                            self.command(
                                "ensight_grpc_shmem.stream_unlock(enscl._shmem_client)",
                                do_eval=False,
                            )

                if self._image_stream is not None:
                    img = self._image_stream.next()
                    buffer = img.pixels

                    while not img.final:
                        img = self._image_stream.next()
                        buffer += img.pixels

                    the_image = dict(pixels=buffer, width=img.width, height=img.height)
                    self._put_image(the_image)
        except Exception:
            # signal that the gRPC connection has broken
            self._image_stream = None
            self._image_thread = None
            self._image = None


class _EnSightSubServicer(ensight_pb2_grpc.EnSightSubscriptionServicer):
    """Internal class handling reverse subscription connections.
    The EnSight gRPC interface has a mechanism for reversing the gRPC
    streams called Subscriptions.  Image and event streams can be
    subscribed to.  In this mode, the client application starts a
    gRPC server that implements the EnSightSubscription protocol.
    EnSight will connect back to the client using this protocol and
    send images/events back to the client as regular (non-stream)
    rpc calls.  This can be useful in situations where it is difficult
    keep a long-running stream alive.
    The EnSightSubServicer class implements a gRPC server for the client application.
    """

    def __init__(self, parent: Optional["EnSightGRPC"] = None):
        self._server: Optional["grpc.Server"] = None
        self._uri: str = ""
        self._parent = parent

    def PublishEvent(self, request: Any, context: Any) -> "ensight_pb2.GenericResponse":
        """Publish an event to the remote server."""
        if self._parent is not None:
            self._parent._put_event(request)
        return ensight_pb2.GenericResponse(str="Event Published")

    def PublishImage(self, request_iterator: Any, context: Any) -> "ensight_pb2.GenericResponse":
        """Publish a single image (possibly in chucks) to the remote server."""
        img: Any = request_iterator.next()
        buffer = img.pixels
        while not img.final:
            img = request_iterator.next()
            buffer += img.pixels
        the_image = dict(pixels=buffer, width=img.width, height=img.height)
        if self._parent is not None:
            self._parent._put_image(the_image)
        return ensight_pb2.GenericResponse(str="Image Published")

    def start(self):
        """Start the gRPC server to be used for the EnSight Subscription Service."""
        self._server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
        ensight_pb2_grpc.add_EnSightSubscriptionServicer_to_server(self, self._server)
        # Start the server on localhost with a random port
        port = self._server.add_insecure_port("localhost:0")
        self._uri = "localhost:" + str(port)
        self._server.start()
